import numpy as np
import xarray as xr
import glob as glob

# from netCDF4 import Dataset
import multiprocessing

# warnings.filterwarnings("ignore")  # TODO don't use normally!!!

def dealiaze(data: xr.Dataset) -> xr.Dataset:
    n_cores = 16  # multiprocessing.cpu_count()
    # if n_cores > 1:
    # 	n_cores = n_cores - 1 # we have the main function running on one core and our institute does not allow to use all cores
    pool = multiprocessing.Pool(n_cores)
    Nyquvist = data.nyquist_velocity.item()  # [6.350626, 4.981658, 2.61496353]
    RangeEnd = data.maximum_range.item()  # [715.4785, 5902.698, 17994.29]
    doppLen = data.doppler_spectrum_length.item() #TODO: check if this is range_vectors or len(range_vectors)
    vel = data.velocity_vectors.values	
    dv = np.diff(vel)[0]
    newVel = np.arange(-Nyquvist * 2, Nyquvist * 2 + dv, dv, dtype=np.float32)
    data["minVelH"] = xr.DataArray(
        dims=("range", "time"), coords={"range": data.range, "time": data.time}
    )
    data["maxVelH"] = data["minVelH"].copy()
    data["minVelV"] = data["minVelH"].copy()
    data["maxVelV"] = data["minVelH"].copy()
    data["minVelZDR"] = data["minVelH"].copy()
    data["maxVelZDR"] = data["minVelH"].copy()
    data["HSpec_rot"] = xr.DataArray(
        dims=("time", "range", "vel_rot"),
        coords={"vel_rot": newVel, "time": data.time, "range": data.range},
    )
    data["VSpec_rot"] = data["HSpec_rot"].copy()
    data["sZDR_rot"] = data["HSpec_rot"].copy()
    # - testing dealiasing:
    for tind, t in enumerate(data.time):
        print(t.values)
        # print(data)
        dataRangeValues = data.range.where(
            np.isfinite(10 * np.log10(data.sel(time=t).ZH)), drop=True
        )
        args = [(r, data.sel(time=t)) for r in dataRangeValues]
        for (
            r,
            HSpec,
            VSpec,
            sZDR,
            sSNR_H,
            sSNR_V,
            maxVelH,
            minVelH,
            maxVelV,
            minVelV,
            maxVelZDR,
            minVelZDR,
            HSpec_rot,
            VSpec_rot,
            sZDR_rot,
        ) in pool.starmap(dealiazeOneHeight, args):
            data["HSpec"].loc[t, r, :] = HSpec
            data["VSpec"].loc[t, r, :] = VSpec
            data["sZDR"].loc[t, r, :] = sZDR
            data["sSNR_H"].loc[t, r, :] = sSNR_H
            data["sSNR_V"].loc[t, r, :] = sSNR_V
            data["maxVelH"].loc[r, t] = maxVelH
            data["minVelH"].loc[r, t] = minVelH
            data["maxVelV"].loc[r, t] = maxVelV
            data["minVelV"].loc[r, t] = minVelV
            data["maxVelZDR"].loc[r, t] = maxVelZDR
            data["minVelZDR"].loc[r, t] = minVelZDR
            data["HSpec_rot"].loc[t, r, :] = HSpec_rot
            data["VSpec_rot"].loc[t, r, :] = VSpec_rot
            data["sZDR_rot"].loc[t, r, :] = sZDR_rot

    return data

def dealiazeOneHeight(r, data):
	sel_range = data.sel(range=r, method='nearest').range.item()	
	idx_sel_range= int(np.arange(len(data.range))[data.range.values == sel_range].item())
	datasel = data.isel(range=slice(idx_sel_range-1, idx_sel_range+2))
	if datasel["chirp_number"].all():
		chirp_number = np.unique(datasel["chirp_number"]).item()
	else:
		raise NotImplementedError('not yet')
	datasel = datasel.sel(chirp=chirp_number).copy()
	maxVel = datasel.nyquist_velocity.item()  # [6.350626, 4.981658, 2.61496353]
	vel = datasel.velocity_vectors.values	
	datasel = datasel.assign_coords(spectrum=vel)
	dv = np.diff(vel)[0]
	ext_vel = np.arange(-maxVel*3,maxVel*3,dv,dtype=np.float32)
	#print(r)
	#datasel = datasel[['sZDR','HSpec','VSpec','sSNR_H','sSNR_V']]
	sZDR = datasel.sZDR
	HSpec = datasel.doppler_spectrum_h
	VSpec = datasel.doppler_spectrum_v
	flattened_h = datasel.doppler_spectrum_h.values.flatten()
	flattened_v = datasel.doppler_spectrum_v.values.flatten()
	flattened_zdr = datasel.sZDR.values.flatten()
	extension_left = 1.75 
	extension_right = extension_left-1.
	
	flattened_h[np.logical_or(ext_vel<-extension_left*maxVel, ext_vel>extension_right*maxVel)] = np.nan	
	flattened_h[flattened_h==0] = np.nan
	flattened_v[np.logical_or(ext_vel<-extension_left*maxVel, ext_vel>extension_right*maxVel)] = np.nan
	flattened_v[flattened_v==0] = np.nan
	flattened_zdr[np.logical_or(ext_vel<-extension_left*maxVel, ext_vel>extension_right*maxVel)] = np.nan
	flattened_v[flattened_zdr==0] = np.nan
	
	import matplotlib.pyplot as plt
	fig, ax = plt.subplots(3, 1, figsize=(10, 10))
	ax[0].plot(ext_vel, flattened_h, label='HSpec')
	ax[0].plot(ext_vel, flattened_v, label='VSpec')
	#vertical line at maxVel
	ax[0].axvline(x=maxVel, color='r', linestyle='--')
	ax[0].axvline(x=-maxVel, color='r', linestyle='--')
	ax[0].axvline(x=0, color='g', linestyle='--')
	ax[0].axvline(x=-extension_left*maxVel, color='b', linestyle='--')
	ax[0].axvline(x=extension_right*maxVel, color='b', linestyle='--')
	
	fig.savefig('extend_spectrum.png')
	# sSNR_H = datasel.sSNR_H
	# sSNR_V = datasel.sSNR_V

	#Get the first non-nan value in flattened_h 
	h_index = np.where(~np.isnan(flattened_h))[0][0]
	v_index = np.where(~np.isnan(flattened_v))[0][0]
	zdr_index = np.where(~np.isnan(flattened_zdr))[0][0]

	#verify if the velocity at index is in the range of -extension_left*maxVel - maxVel
	aliazing_flag = False
	for index_ in [h_index, v_index, zdr_index]:
		aliazing_flag = aliazing_flag or (ext_vel[index_] < -maxVel and ext_vel[index_] > -extension_left*maxVel)		
	
	breakpoint()
	if (~np.isnan(sZDR.sel(spectrum=maxVel,method='nearest')) or ~np.isnan(sZDR.sel(spectrum=-maxVel,method='nearest')) or ~np.isnan(sZDR.sel(spectrum=-maxVel+dv,method='nearest')) 
		or ~np.isnan(sZDR.sel(spectrum=maxVel-dv,method='nearest')) or ~np.isnan(sZDR.sel(spectrum=-maxVel+2*dv,method='nearest')) or ~np.isnan(sZDR.sel(spectrum=maxVel-2*dv,method='nearest')) 
		or ~np.isnan(sZDR.sel(spectrum=-maxVel+3*dv,method='nearest')) or ~np.isnan(sZDR.sel(spectrum=maxVel-3*dv,method='nearest'))): 
	# if we have non-noise values at +- Nyquvist range or upto 3 vel bins to the right or left
		datadrop = datasel.where(data.Vel < maxVel,drop=True) #- since we regridded, I need to drop everything that is larger or smaller than +- Ny to actually kit the spectra together (otherwise we would have a lot of nan in between..)
		datadrop = datadrop.where(data.Vel > -maxVel,drop=True)
		q = len(datadrop.Vel.values)-1
		k = 0
		
		if np.isnan(datadrop.HSpec.isel(Vel=q)):
			datadrop.HSpec.values[q] = -99999
		for i in range(q):
			if np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+1])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+2])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+3])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+4])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+5])):
				datadrop.HSpec.values[i] = -99999
		if datadrop.HSpec.isnull().any(dim='Vel'):
			while ~np.isnan(datadrop.HSpec.isel(Vel=q)): #-- find out how much we need to shift
				k = k + 1
				q = q - 1
		# now roll the coordinates:
		datadrop = datadrop.where(datadrop.HSpec!=-99999,np.nan)
		#print(maxVel)
		dataRoll = datadrop.roll(Vel=k+1,roll_coords=False)  # shift everything, the direction doesnt matter
		
		dataRollRegridded = dataRoll.reindex({'Vel':data.Vel.values}, method = 'nearest', tolerance = 0.05) # now I need to go back to original velocity
		HSpec = dataRollRegridded.HSpec
		VSpec = dataRollRegridded.VSpec
		sZDR = dataRollRegridded.sZDR
		# sSNR_H = dataRollRegridded.sSNR_H
		# sSNR_V = dataRollRegridded.sSNR_V
	
	
	#- now move everything to 0 and name that rot (better for plotting)
	velMatrix = HSpec.values/HSpec.values*vel
	maxVelH = np.nanmax(velMatrix)#,axis=1)
	minVelH = np.nanmin(velMatrix)#,axis=1)
	if np.isnan(maxVelH):
		maxVelH = 0
	if np.isnan(minVelH):
		minVelH = 0
	
	velMatrix = VSpec.values/VSpec.values*vel
	maxVelV = np.nanmax(velMatrix)#,axis=1)
	minVelV = np.nanmin(velMatrix)#,axis=1)
	if np.isnan(maxVelV):
		maxVelV = 0
	if np.isnan(minVelV):
		minVelV = 0
		
	velMatrix = sZDR.values/sZDR.values*vel
	maxVelZDR = np.nanmax(velMatrix)#,axis=1)
	minVelZDR = np.nanmin(velMatrix)#,axis=1)
	if np.isnan(maxVelZDR):
		maxVelZDR = 0
	if np.isnan(minVelZDR):
		minVelZDR = 0
	#- now shift everything to 0, because we might get folding I have to increase Dopplervelocity range to have +-12
	HSpecRe = HSpec.reindex({'Vel':newVel}, method = 'nearest', tolerance = 0.05) 
	vel2rot = int((-maxVelH/dv).round())
	if datasel.maxVelV > 0:
		HSpec_rot = HSpecRe.roll(Vel=vel2rot-1,roll_coords=False)
	else:
		HSpec_rot = HSpecRe.roll(Vel=vel2rot,roll_coords=False)
	
	VSpecRe = VSpec.reindex({'Vel':newVel}, method = 'nearest', tolerance = 0.05) 	
	vel2rot = int((-maxVelV/dv).round())
	if datasel.maxVelV > 0:
		VSpec_rot = VSpecRe.roll(Vel=vel2rot-1,roll_coords=False)
	else:
		VSpec_rot = VSpecRe.roll(Vel=vel2rot,roll_coords=False)
		
	sZDRRe = sZDR.reindex({'Vel':newVel}, method = 'nearest', tolerance = 0.05) 
	vel2rot = int((-maxVelZDR/dv).round())
	if datasel.maxVelZDR > 0:
		sZDR_rot = sZDRRe.roll(Vel=vel2rot-1,roll_coords=False)
	else:
		sZDR_rot = sZDRRe.roll(Vel=vel2rot,roll_coords=False)
	sSNR_H,sSNR_V = None, None #TODO
	return r,HSpec,VSpec,sZDR,sSNR_H,sSNR_V,maxVelH,minVelH,maxVelV,minVelV,maxVelZDR,minVelZDR,HSpec_rot,VSpec_rot,sZDR_rot



from datetime import datetime
from pathlib import Path
import xarray as xr 

from matplotlib import pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.figure import Figure
import numpy as np

from gfatpy.radar.rpg_nc import rpg
from gfatpy.utils.utils import parse_datetime


class nebula():
    def __init__(self, ka_nc_path: Path, ww_nc_path: Path):
        self.ka = rpg(ka_nc_path)
        self.ww = rpg(ww_nc_path)        
        self._level = None
        self.nb = self.ka
        self.nb._data = None

    @property
    def level(self) -> int:
        if self.ka.level != self.ww.level:
            raise ValueError("Ka- and W-band measurement levels do not match.")        
        self._level = self.ka.level
        return self._level

    @property
    def data(self) -> xr.Dataset:
        if self.level == 1:
            if self.nb._data is None:
                _data = self.ka.data.copy()
                _data["DWR"] = 10 * np.log10(self.ka.data["Ze"] / self.ww.data["Ze"])
                _data["DWR"].attrs["units"] = "dB"
                _data["DWR"].attrs["long_name"] = "Ka-W DWR"

                _data["DDV"] = self.ka.data["v"] - self.ww.data["v"]
                _data["DDV"].attrs["units"] = "m/s"
                _data["DDV"].attrs["long_name"] = "Ka-W DDV"
            else:
                _data = self.nb._data
        return _data

    def quicklook(self, variable: str | list[str] | None = None, **kwargs):
        self.ka.quicklook(variable, **kwargs)
        self.ww.quicklook(variable, **kwargs)

    def plot_profile(
        self,
        target_time: datetime | np.datetime64,
        range_limits: tuple[float, float],
        variable: str | list[str] | None = None,
        **kwargs,
    ) -> tuple[Figure, Path | None]:
        _kwargs = kwargs.copy()        
        _kwargs["savefig"] = False
        _kwargs["color_list"] = [kwargs.get("ka_color", "red")]

        if variable is None:
            variables = [str(var) for var in self.ka.data.variables.keys()]
        if isinstance(variable, str):
            variables = [variable]
        elif isinstance(variable, list):
            variables = variable

        for variable_ in variables:
            fig1, _ = self.ka.plot_profile(
                target_times=target_time,
                range_limits=range_limits,
                variable=variable_,
                **_kwargs,
            )

            _kwargs["fig"] = fig1
            _kwargs["color_list"] = [kwargs.get("w_color", "blue")]
            
            fig, _ = self.ww.plot_profile(
                target_times=target_time,
                range_limits=range_limits,
                variable=variable_,
                **_kwargs,
            )

            ka_line, w_line = fig.findobj(Line2D)[0], fig.findobj(Line2D)[1]
            ka_line.set_label(f"Ka-band")
            w_line.set_label(f"W-band")

            ax = fig.get_axes()[0]
            ax.legend()

            if kwargs.get("savefig", False):
                filepath = (
                    kwargs.get("output_dir", Path.cwd())
                    / f"profile_{variable_}_{target_time:%Y%m%dT%H%M%S}.png"
                )
                fig.savefig(filepath, dpi=kwargs.get("dpi", 300))
        return fig, filepath

    def plot_timeseries(
        self,
        target_range: float,
        time_limits: tuple[datetime | np.datetime64, datetime | np.datetime64] | None = None,
        variable: str | None = None,
        **kwargs,
    ) -> tuple[Figure, Path | None]:
        
        _kwargs = kwargs.copy()        
        _kwargs["savefig"] = False
        _kwargs["color_list"] = [kwargs.get("ka_color", "red")]

        if  time_limits is None:
            time_limits = (parse_datetime(self.ka.data.time[0].values), parse_datetime(self.ka.data.time[-1].values))

        fig1, _ = self.ka.plot_timeseries(
            target_ranges=target_range,
            time_limits=time_limits,
            variable=variable,
            **_kwargs,
        )

        _kwargs["fig"] = fig1
        _kwargs["color_list"] = [kwargs.get("w_color", "blue")]
        fig, _ = self.ww.plot_timeseries(
            target_ranges=target_range,
            time_limits=time_limits,
            variable=variable,
            **_kwargs,
        )

        ka_line, w_line = fig.findobj(Line2D)[0], fig.findobj(Line2D)[1]
        ka_line.set_label(f"Ka-band")
        w_line.set_label(f"W-band")

        ax = fig.get_axes()[0]
        variable_string = self.ka.data[variable].attrs['long_name'].replace('W-band','').replace("Ka-band",'')
        ax.set_ylabel(f"{variable_string}, [{self.ka.data[variable].attrs['units']}]")
        ax.legend()

        if kwargs.get("savefig", False):
            filepath = (
                kwargs.get("output_dir", Path.cwd())
                / f"timeseries_{variable}_{time_limits[0]:%Y%m%dT%H%M%S}_{time_limits[-1]:%Y%m%dT%H%M%S}_{target_range}.png"
            )
            fig.savefig(filepath, dpi=kwargs.get("dpi", 300))
        return fig, filepath


    def plot_spectrum(
        self, target_time: datetime | np.datetime64, target_range: float, **kwargs
    ) -> tuple[Figure, Path | None]:
        _kwargs = kwargs.copy()
        _kwargs.pop("output_dir")
        _kwargs.pop("savefig")
        _kwargs["color"] = kwargs.get("ka_color", "red")
        fig, _ = self.ka.plot_spectrum(target_time, target_range, **_kwargs)

        _kwargs["fig"] = fig
        _kwargs["color"] = kwargs.get("w_color", "blue")
        fig, _ = self.ww.plot_spectrum(target_time, target_range, **_kwargs)
        ka_line, w_line = fig.findobj(Line2D)[0], fig.findobj(Line2D)[2]
        ka_line.set_label(f"Ka-band {ka_line.get_label()}")
        w_line.set_label(f"W-band {w_line.get_label()}")

        ax = fig.get_axes()[0]
        ax.legend()

        if kwargs.get("savefig", False):
            filepath = (
                kwargs.get("output_dir", Path.cwd())
                / f"{target_time:%Y%m%dT%H%M%S}_{target_range:.0f}_spectrum.png"
            )
            fig.savefig(filepath, dpi=kwargs.get("dpi", 300))

        return fig, filepath



import os
from typing import Tuple
import numpy as np
import pandas as pd
from pathlib import Path
from datetime import datetime
from matplotlib.dates import DateFormatter
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import matplotlib.pyplot as plt


from gfatpy.aeronet.utils import AERONET_INFO
from gfatpy.utils.plot import color_list

# FIXME: type hinting
def distribution(
    df: pd.DataFrame,
    date_string: list[str] | str,
    type_distribution: str = "volume",
    figure_dir: Path | str | None = None,
    xlims: Tuple[float, float] | str = "auto",
    ylims: Tuple[float, float] | str = "auto",
    dpi: int = 400,
    fig_size: Tuple[float, float] = (8, 6),
) -> Tuple[Figure, plt.Axes, str] | Tuple[list, list, list]:  # type: ignore
    """Distribution() plots all the volume/surface/number size distributions on the date 'date_string'
    
    Args:

        - df (pd.DataFrame): AERONET dataframe loaded with `aeronet.reader`.
        - date_string (str | list[str]): date string or list of date strings in format '%Y-%m-%d'.
        - type_distribution (str [optional]): type of distribution 'volume', 'surface', 'number'
        - figure_dir (str [optional]): file path to save the figure. Directory must exist.
        - xlims (list [optional]): 2D-array with x-axis limits.
        - ylims (list [optional]): 2D-array with y-axis limits.
        - clims (list [optional]): 2D-array with z-axis limits.
        - dpi (int, optional): dots-per-inch figure (higher means larger quality and file size)
    
    Returns:

        - figure handle (matplotlib.figure.Figure): handle of figure | list of figure handles
        - axes handles (matplotlib.axes.Axes): handle of axis | list of axis handles
        - Figure file (str): figure file path | list of figure file paths
    """

    if type(date_string) is str:
        date_string_list = [date_string]
    elif type(date_string) is list:
        date_string_list = date_string
    else:
        raise RuntimeError("Wrong type of `date_string`.")

    if isinstance(figure_dir, str):
        figure_dir = Path(figure_dir)
        if figure_dir.is_dir() is False:
            raise RuntimeError("`figure_dir` not found.")
    elif figure_dir is None:
        figure_dir = Path(".")

    fig_list, axes_list, fig_pathname_list = [], [], []

    df["date"] = [pdate.date() for pdate in df.index.to_pydatetime()]  # type: ignore
    df["time"] = [pdate.time() for pdate in df.index.to_pydatetime()]  # type: ignore

    # min and max column index of the volumne size distribution
    idx_min = df.columns.get_loc("0.050000")
    idx_max = df.columns.get_loc("15.000000")

    # find maximum value
    if ylims == "auto":
        time_condition = np.ones(len(df)) * False
        for date_ in date_string_list:
            time_condition_ = df["date"] == datetime.strptime(date_, "%Y-%m-%d").date()
            time_condition = np.logical_or(time_condition, time_condition_)
        max_dV_dlnr = df.iloc[time_condition.values, np.arange(idx_min, idx_max + 1, 1)].max().max()  # type: ignore # VSD um^3/um^2

        # Found the order of magnitude of the maximum value to round the ylims
        order = np.floor(np.log10(max_dV_dlnr))
        round_max_value: float = np.ceil(max_dV_dlnr / np.power(10, order)) * np.power(
            10, order
        )
        ylims = (0.0, round_max_value)

    for date_ in date_string_list:
        time_condition = df["date"] == datetime.strptime(date_, "%Y-%m-%d").date()
        # Find radius in df header
        radius = np.asarray(AERONET_INFO["radius"])
        df_ = df.loc[time_condition]  # VSD um^3/um^2
        if df_.empty:
            raise RuntimeError(f"No data on {date_}")
        dV_dlnr = df_.iloc[:, np.arange(idx_min, idx_max + 1, 1)]  # VSD um^3/um^2
        if 'Site' in df_.columns:
            site_str = df_["Site"][0]
        else:
            site_str = df_["AERONET_Site"][0]
        if len(dV_dlnr) > 0:
            # Figure
            fig, axes = plt.subplots(1, figsize=fig_size)
            # fig.set_figwidth(7)
            # fig.set_figheight(5)
            colors = color_list(len(dV_dlnr))
            for i_ in np.arange(len(dV_dlnr)):
                if type_distribution == "volume":
                    axes.plot(
                        radius,
                        dV_dlnr.iloc[i_, :],
                        label=str(df_["time"][i_]),
                        c=colors[i_],
                        linewidth=3,
                    )  # type: ignore
                    axes.set_ylabel(r"dV/dlnr, [$\mu$$m^3$/$\mu$$m^2$]")
                elif type_distribution == "surface":
                    Vradius = (4.0 / 3.0) * np.pi * np.power(radius, 3)  # i-bin volume
                    dN_dlnr = dV_dlnr.iloc[i_, :] / Vradius  # NSD #/um^2 # type: ignore
                    Sradius = 4.0 * np.pi * np.power(radius, 2)  # i-bin surface
                    # dSlnr_dlnr = (3 * dV_dlnr.iloc[i_,:] / radius ) #SSD #um^2/um^2
                    dS_dlnr = dN_dlnr * Sradius
                    axes.plot(
                        radius,
                        dS_dlnr,
                        label=df_["time"][i_],
                        c=colors[i_],
                        linewidth=3,
                    )
                    axes.set_ylabel(r"dS/dlnr, [$\mu$$m^2$/$\mu$$m^2$]")
                elif type_distribution == "number":
                    Vradius = (4.0 / 3.0) * np.pi * np.power(radius, 3)  # i-bin volume
                    dNlnr_dlnr = (
                        dV_dlnr.iloc[i_, :] / Vradius
                    )  # NSD #/um^2 # type: ignore
                    axes.plot(
                        radius,
                        dNlnr_dlnr,
                        label=df_["time"][i_],
                        c=colors[i_],
                        linewidth=3,
                    )
                    axes.set_ylabel(r"dN/dlnr, [#/$\mu$$m^2$]")
            axes.set_title(f"{site_str} | {date_}")
            axes.set_xlabel(r"radius, [$\mu$m]")
            axes.set_xscale("log")

            if isinstance(xlims, tuple) and len(xlims) > 0:
                axes.set_xlim(*xlims)
            if isinstance(ylims, tuple) and len(ylims) > 0:
                axes.set_ylim(*ylims)
            axes.legend(loc="best")
            axes.grid(axis="y")

            filename = os.path.join(
                figure_dir, f"{type_distribution}_concentration_{date_}.png"
            )
            plt.savefig(filename, dpi=dpi, bbox_inches="tight")
            fig_list.append(fig)
            axes_list.append(axes)
            fig_pathname_list.append(filename)
    if len(date_string_list) == 1:
        fig_output = fig_list[0]
        axes_output = axes_list[0]
        fig_pathname_output = fig_pathname_list[0]
    else:
        fig_output = fig_list
        axes_output = axes_list
        fig_pathname_output = fig_pathname_list

    return fig_output, axes_output, fig_pathname_output  # type: ignore


def aod(
    df: pd.DataFrame,
    header: dict,
    properties2plot: list[str],
    initial_date: datetime,
    final_date: datetime,
    figure_dir: Path | str,
    xlims: Tuple[float, float] | str = "auto",
    ylims: Tuple[float, float] | str = "auto",
    dpi: int = 400,
) -> Tuple[Figure, Axes, str]:
    """aod() plots all the volume/surface/number size distributions on the date 'date_string'
    
    Args:

        - df (pd.DataFrame): AERONET dataframe loaded from reader_all.
        - header (dict): header of AERONET file.
        - initial_date (str): date string in format '%Y-%m-%d'.
        - final_date (str): date string in format '%Y-%m-%d'.
        - figure_dir (str [optional]): file path to save the figure. Directory must exist.
        - xlims (list [optional]): 2D-array with x-axis limits.
        - ylims (list [optional]): 2D-array with y-axis limits.
        - dpi (int, optional): dots-per-inch figure (higher means larger quality and file size)
    
    Returns:
    
        - figure handle (matplotlib.figure.Figure): handle of figure
        - axes handles (matplotlib.axes.Axes): handle of axis
        - Figure file (str): figure file path
    """

    site_str = header["aeronet_station"]

    if figure_dir is None:
        figure_dir = Path(".")
    elif isinstance(figure_dir, str):
        figure_dir = Path(figure_dir)

    if not figure_dir.exists() or not figure_dir.is_dir():
        raise RuntimeError("`figure_dir` not found.")

    if initial_date > final_date:
        raise RuntimeError("`initial_date` should be lower than `final_date`.")

    df["date"] = [pdate.date() for pdate in df.index.to_pydatetime()]  # type: ignore
    df["time"] = [pdate.time() for pdate in df.index.to_pydatetime()]  # type: ignore

    raws = df.index[
        np.logical_and(
            df["date"] >= initial_date.date(), df["date"] <= final_date.date()
        )
    ]
    columns = properties2plot

    fig, axes = plt.subplots(1, 1, figsize=[10, 5])

    colors = np.flip(color_list(len(columns)), axis=0)

    for idx, property in enumerate(properties2plot):
        df.loc[raws, property].plot(
            marker="o", markersize=5, c=colors[idx, :], linewidth=0, ax=axes
        )
    axes.set_ylabel("AOD, [#]")
    if initial_date == final_date:
        date2show = datetime.strftime(initial_date, "%d/%m/%Y")
    else:
        if initial_date.year == final_date.year:
            if initial_date.month == final_date.month:
                date2show = f"{datetime.strftime(initial_date, '%d')}-{datetime.strftime(final_date, '%d/%m/%Y')}"
            else:
                date2show = f"{datetime.strftime(initial_date, '%d/%m')}-{datetime.strftime(final_date, '%d/%m/%Y')}"
        else:
            date2show = f"{datetime.strftime(initial_date, '%d/%m/%Y')}-{datetime.strftime(final_date, '%d/%m/%Y')}"
    axes.set_title(f"{site_str} | {date2show}")

    if (final_date - initial_date).days < 3:
        myFmt = DateFormatter("%H %d-%m-%y")
        axes.set_xlabel(r"Datetime, [hour day-month-year]")
    elif (final_date - initial_date).days > 3 and (final_date - initial_date).days < 7:
        myFmt = DateFormatter("%d-%m-%y")
        axes.set_xlabel(r"Date, [day-month-year]")
    else:
        myFmt = DateFormatter("%m-%y")
        axes.set_xlabel(r"Date, [month-year]")
    axes.xaxis.set_major_formatter(myFmt)

    if isinstance(xlims, tuple) and len(xlims) > 0:
        axes.set_xlim(*xlims)

    if ylims == "auto":
        max_aod = df.loc[raws, properties2plot].max().max()
        ylims = (0.0, np.round(max_aod * 1.1, 2))

    if isinstance(ylims, tuple) and len(ylims) > 0:
        axes.set_ylim(*ylims)

    axes.legend(loc="best")
    axes.grid()

    if initial_date == final_date:
        figure_pathname = os.path.join(
            figure_dir, f"aod_{str(initial_date.date())}.png"
        )
    else:
        figure_pathname = os.path.join(
            figure_dir, f"aod_{str(initial_date.date())}-{str(final_date.date())}.png"
        )

    plt.savefig(figure_pathname, dpi=dpi, bbox_inches="tight")
    return fig, axes, figure_pathname


# def plot_single_scatter(
#     df,
#     cf_info,
#     wavelength=440,
#     min_radius=50,
#     colorbar_property="Angstrom_Exponent_440-870nm_from_Coincident_Input_AOD",
#     filter_name="",
#     xlims=[],
#     ylims=[],
#     clims=[],
# ):
#     """
#     apc_scatter makes scatter plots of up to three dataframes readed with function reader_all.
#     Input:
#     wavelength: AOD wavelength in nm (e.g., 440)
#     minimum radius: minimum radius to integrate the NSD in nm (e.g., 50)
#     colorbar_property: variable in dataframe for colorbar (e.g., 'Angstrom_Exponent_440-870nm_from_Coincident_Input_AOD' [default])
#     filter_name: Allows to indicate in the figure title in the data are filtered.
#     xlims: set x limits
#     ylims: set y limits
#     clims: set colorbar limits
#     """

#     # Define colorbar label
#     colorbar_label_dict = {
#         "Angstrom_Exponent_440-870nm_from_Coincident_Input_AOD": "AE[440-870nm]",
#         "Sphericity_Factor(%)": "Sphericity Factor(%)",
#         "Depolarization_Ratio[440nm]": "Depol. ratio[440nm]",
#         "Depolarization_Ratio[675nm]": "Depol. ratio[675nm]",
#         "FMF[675nm]": "FMF][675nm]",
#         "Lidar_Ratio[440nm]": "Lidar ratio[440nm]",
#         "Single_Scattering_Albedo[440nm]": "SSA[440nm]",
#     }

#     if not (colorbar_property in colorbar_label_dict.keys()):
#         colorbar_label_dict[colorbar_property] = colorbar_property

#     # Define y label
#     lnAOD = "lnAOD%d" % wavelength
#     lnN = "ln_n%d" % min_radius
#     ylabel_string = "$ln(N_{%d})$" % min_radius
#     xlabel_string = "$ln(AOD[%d nm])$" % wavelength

#     # Figure
#     fig, axes = plt.subplots(1, 1)
#     fig.set_figwidth(10)
#     fig.set_figheight(6)

#     # Scatter
#     ax_mappable = axes.scatter(df[lnAOD], df[lnN], c=df[colorbar_property])

#     # Linear fit
#     exponent, exponent_error, intercept, intercept_error, R2 = cf_info
#     m, error_m, n, error_n, R2 = (
#         exponent,
#         exponent_error,
#         np.log(intercept),
#         intercept_error / intercept,
#         R2,
#     )
#     plt.plot(
#         df[lnAOD],
#         m * df[lnAOD] + n,
#         "r",
#         label="y=(%.4f $\pm$ %.4f)x + (%.4f $\pm$ %.4f) [R=%.2f]"
#         % (m, error_m, n, error_n, np.sqrt(R2)),
#     )

#     # Axes format
#     if len(filter_name) > 0:
#         axes.set_title(
#             "%s | Filter: %s | data[#]: %d " % (df["Site"][0], filter_name, len(df))
#         )
#     else:
#         axes.set_title("%s | data[#]: %d " % (df["Site"][0], len(df)))
#     axes.legend()
#     axes.set_xlabel(xlabel_string)
#     axes.set_ylabel(ylabel_string)
#     if len(xlims) > 0:
#         ax_mappable.axes.set_xlim(xlims)
#     if len(ylims) > 0:
#         ax_mappable.axes.set_ylim(ylims)
#     if len(clims) > 0:
#         ax_mappable.set_clim(clims)
#     fig.colorbar(ax_mappable, label=colorbar_label_dict[colorbar_property], ax=axes)
#     fig.tight_layout()
#     return fig, axes


# def plot_scatter(
#     df_tuple,
#     x="lnAOD440",
#     y="ln_n50",
#     colorbar_property="Angstrom_Exponent_440-870nm_from_Coincident_Input_AOD",
#     filter_name="",
#     xlims=[],
#     ylims=[],
#     clims=[],
# ):
#     """
#     apc_scatter makes scatter plots of up to three dataframes readed with function reader_all.
#     Input:
#     x: variable in dataframe (e.g., 'Depolarization_Ratio[440nm]')
#     y: variable in dataframe (e.g., 'Lidar_Ratio[440nm]')
#     colorbar_property: variable in dataframe for colorbar (e.g., 'Angstrom_Exponent_440-870nm_from_Coincident_Input_AOD' [default])
#     filter_name: Allows to indicate in the figure title in the data are filtered.
#     xlims: set x limits
#     ylims: set y limits
#     clims: set colorbar limits
#     """
#     SMALL_SIZE = 12
#     MEDIUM_SIZE = 14
#     BIGGER_SIZE = 18

#     plt.rc("font", size=MEDIUM_SIZE)  # controls default text sizes
#     plt.rc("axes", titlesize=MEDIUM_SIZE)  # fontsize of the axes title
#     plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
#     plt.rc("xtick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
#     plt.rc("ytick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
#     plt.rc("legend", fontsize=MEDIUM_SIZE)  # legend fontsize
#     plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title
#     # Define colorbar label
#     colorbar_label_dict = {
#         "Angstrom_Exponent_440-870nm_from_Coincident_Input_AOD": "AE[440-870nm]",
#         "Sphericity_Factor(%)": "Sphericity Factor(%)",
#         "Depolarization_Ratio[440nm]": "Depol. ratio[440nm]",
#         "Depolarization_Ratio[675nm]": "Depol. ratio[675nm]",
#         "FMF[675nm]": "FMF][675nm]",
#         "Lidar_Ratio[440nm]": "Lidar ratio[440nm]",
#         "Single_Scattering_Albedo[440nm]": "SSA[440nm]",
#     }

#     if not (colorbar_property in colorbar_label_dict.keys()):
#         colorbar_label_dict[colorbar_property] = colorbar_property

#     # Define y label
#     ylabel_dic = {
#         "lnN_fine": "$ln(N_{fine})$",
#         "lnN_coarse": "$ln(N_{coarse}$)",
#         "lnN": "$ln(N)$",
#     }
#     if not y in ylabel_dic.keys():
#         ylabel_dic[y] = y

#     xlabel_dic = {"lnAOD440": "$ln(AOD[440nm])$"}
#     if not x in xlabel_dic.keys():
#         xlabel_dic[x] = x

#     # Figure
#     fig, axes = plt.subplots(1, len(df_tuple))
#     fig.set_figwidth(15)
#     fig.set_figheight(3)

#     for axes_idx in np.arange(len(df_tuple)):
#         df = df_tuple[axes_idx]
#         # Scatter
#         ax_mappable = axes[axes_idx].scatter(
#             df[x], df[y], c=df[colorbar_property], cmap="YlOrRd"
#         )

#         # Linear fit
#         m, error_m, n, error_n, R2 = linrest(df[x], df[y])
#         axes[axes_idx].plot(
#             df[x],
#             m * df[x] + n,
#             "r",
#             lw=3,
#             label="AERONET | R=%.2f | m=%.3f$\pm$%.3f | n=%.3f$\pm$%.3f"
#             % (np.sqrt(R2), m, error_m, n, error_n),
#         )
#         # Axes format
#         if len(filter_name) > 0:
#             # axes[axes_idx].set_title('%s | data[#]: %d | %s\n R=%.2f | m=%.3f$\pm$%.3f | n=%.3f$\pm$%.3f' % (filter_name, len(df), df['Site'][0], np.sqrt(R2), m, error_m, n, error_n))
#             axes[axes_idx].set_title(
#                 "%s | data[#]: %d | %s" % (filter_name, len(df), df["Site"][0])
#             )
#         else:
#             axes[axes_idx].set_title("%s" % (df["Site"][0]))
#             # axes[axes_idx].set_title('%s\n R=%.2f | m=%.2f$\pm$%.2f | n=%.2f$\pm$%.2f' % (df['Site'][0], np.sqrt(R2), m, error_m, n, error_n))
#             # axes[axes_idx].set_title('%s\n $R$=%.2f |alpha=%.2f | C=%.2f' % (df['Site'][0], R, m, np.exp(n)))
#         axes[axes_idx].set_facecolor("gainsboro")
#         axes[axes_idx].set_xlabel(xlabel_dic[x])
#         axes[axes_idx].legend()
#         if axes_idx == 0:
#             axes[axes_idx].set_ylabel(ylabel_dic[y])

#         if len(clims) > 1:
#             ax_mappable.set_clim(clims)
#         else:
#             if axes_idx == 0:
#                 clims = ax_mappable.get_clim()
#         if len(xlims) > 0:
#             ax_mappable.axes.set_xlim(xlims)
#         if len(ylims) > 0:
#             ax_mappable.axes.set_ylim(ylims)
#         fig.colorbar(
#             ax_mappable, label=colorbar_label_dict[colorbar_property], ax=axes[axes_idx]
#         )
#     return fig, axes



import datetime
import logging
import copy
import os
import collections

import numpy as np
import pytz

from .generic import BaseLidarMeasurement, LidarChannel
from .diva import DivaConverterMixin

logger = logging.getLogger(__name__)

c = 299792458.0  # Speed of light


class LicelChannelData:
    """ A class representing a single channel found in a single Licel file."""

    def __init__(self, raw_info, raw_data, duration, use_id_as_name=False, channel_name=None):
        """
        This is run when creating a new object.

        Parameters
        ----------
        raw_info : dict
           A dictionary containing raw channel information.
        raw_data : dict
           An array with raw channel data.
        duration : float
           Duration of the file, in seconds
        use_id_as_name : bool
           If True, the transient digitizer name (e.g. BT0) is used as a channel
           name. If False, a more descriptive name is used (e.g. '01064.o_an').
        channel_name : str or None
           If provided, it will override the automatic generated channel name. It can be used if names are not unique.
        """
        self.raw_info = raw_info
        self.raw_data = raw_data
        self.duration = duration
        self.use_id_as_name = use_id_as_name
        self.channel_name_input = channel_name
        self._assign_properties()

    def _assign_properties(self):
        """ Assign properties """
        self.adcbits = int(self.raw_info['ADCbits'])
        self.active = int(self.raw_info['active'])
        self.analog_photon = self.raw_info['analog_photon']
        self.bin_width = float(self.raw_info['bin_width'])
        self.data_points = int(self.raw_info['number_of_datapoints'])
        self.hv = float(self.raw_info['HV'])
        self.id = self.raw_info['ID']
        self.laser_used = int(self.raw_info['laser_used'])
        self.number_of_shots = int(self.raw_info['number_of_shots'])
        self.wavelength_str = self.raw_info['wavelength']

        self.address = int(self.id[-1:], base=16)

        if self.is_analog:
            self.discriminator = float(self.raw_info['discriminator']) * 1000  # Analog range in mV
        else:
            self.discriminator = float(self.raw_info['discriminator'])

    @property
    def is_photodiode(self):
        return self.id[0:2] == 'PD'

    @property
    def wavelength(self):
        """ Property describing the nominal wavelength of the channel.

        Returns
        -------
        : int or None
           The integer value describing the wavelength. If no raw_info have been provided,
           returns None.
        """
        wavelength = self.wavelength_str.split('.')[0]
        return int(wavelength)

    @property
    def channel_name(self):
        """
        Construct the channel name adding analog photon info to avoid duplicates

        If use_id_as_name is True, the channel name will be the transient digitizer ID (e.g. BT01).
        This could be useful if the lidar system has multiple telescopes, so the descriptive name is
        not unique.

        Returns
        -------
        channel_name : str
           The channel name
        """
        if self.channel_name_input is not None:
            return self.channel_name_input

        if self.use_id_as_name:
            channel_name = self.id
        else:
            acquisition_type = self.analog_photon_string
            channel_name = "%s_%s" % (self.wavelength_str, acquisition_type)
        return channel_name

    @property
    def analog_photon_string(self):
        """ Convert the analog/photon flag found in the Licel file to a proper sting.

        Returns
        -------
        string : str
           'an' or 'ph' string, for analog or photon-counting channel respectively.
        """
        if self.analog_photon == '0':
            string = 'an'
        elif self.analog_photon == '1':
            string = 'ph'
        elif self.analog_photon == '2':
            string = 'std_an'
        elif self.analog_photon == '3':
            string = 'std_ph'
        else:
            string = str(self.analog_photon)
        return string

    def calculate_physical(self):
        """ Calculate physically-meaningful data from raw channel data:

        * In case of analog signals, the data are converted to mV.
        * In case of photon counting signals, data are stored as number of photons.

        In addition, some ancillary variables are also calculated (z, dz, number_of_bins).
        """

        norm = self.raw_data / float(self.number_of_shots)
        dz = self.bin_width

        if self.is_analog:
            # If the channel is in analog mode
            ADCrange = self.discriminator  # Discriminator value already in mV

            if self.is_photodiode and (self.adcbits == 0):
                logger.debug(
                    "Assuming adcbits equal 1. This is a bug in current licel format when storing photodiode data.")
                channel_data = norm * ADCrange / (2 ** self.adcbits)
            else:
                channel_data = norm * ADCrange / (
                            (2 ** self.adcbits) - 1)  # Licel LabView code has a bug (does not account -1).

        else:
            channel_data = norm * self.number_of_shots

        # Calculate Z
        self.z = np.array([dz * bin_number + dz / 2.0 for bin_number in range(self.data_points)])
        self.dz = dz
        self.data = channel_data

    @property
    def is_analog(self):
        return self.analog_photon == '0'

    @property
    def laser_shots(self):
        """ Alias for number_of_shots """
        return self.number_of_shots


class LicelFile(object):
    """ A class representing a single binary Licel file. """

    licel_file_header_format = ['filename',
                                'start_date start_time end_date end_time altitude longitude latitude zenith_angle',
                                # Appart from Site that is read manually
                                'LS1 rate_1 LS2 rate_2 number_of_datasets', ]
    licel_file_channel_format = 'active analog_photon laser_used number_of_datapoints 1 HV bin_width wavelength d1 d2 d3 d4 ADCbits number_of_shots discriminator ID'

    channel_data_class = LicelChannelData

    # If True, it corrects the old Raymetrics convention of zenith angle definition (zenith = -90 degrees)
    fix_zenith_angle = False

    def __init__(self, file_path, use_id_as_name=False, get_name_by_order=False, licel_timezone="UTC", import_now=True):
        """
        This is run when creating a new object.

        Parameters
        ----------
        file_path : str
           The path to the Licel file.
        use_id_as_name : bool
           If True, the transient digitizer name (e.g. BT0) is used as a channel
           name. If False, a more descriptive name is used (e.g. '01064.o_an').
        get_name_by_order : bool
           If True, the channel name is given by the order of the channel in the file. In this case the
           `use_id_as_name` variable is ignored.
        licel_timezone : str
           The timezone of dates found in the Licel files. Should match the available
           timezones in the TZ database.
        import_now : bool
           If True, the header and data are read immediately. If not, the user has to call the
           corresponding methods directly. This is used to speed up reading files when only
           header information are required.
        """
        self.file_path = file_path
        self.file_name = os.path.basename(file_path)

        self.use_id_as_name = use_id_as_name
        self.get_name_by_order = get_name_by_order
        self.start_time = None
        self.stop_time = None
        self.licel_timezone = licel_timezone

        self.header_lines = []  # Store raw header lines, to be used in save_as_txt

        if import_now:
            self.import_file()
        else:
            self.import_header_only()

    def import_file(self):
        """ Read the header info and data of the Licel file.
        """
        channels = collections.OrderedDict()
        photodiodes = collections.OrderedDict()

        with open(self.file_path, 'rb') as f:

            self.read_header(f)

            # Check the complete header is read
            f.readline()

            # Import the data
            for channel_no, current_channel_info in enumerate(self.channel_info):
                raw_data = np.fromfile(f, 'i4', int(current_channel_info['number_of_datapoints']))
                a = np.fromfile(f, 'b', 1)
                b = np.fromfile(f, 'b', 1)

                if (a[0] != 13) | (b[0] != 10):
                    logger.warning("No end of line found after record. File could be corrupt: %s" % self.file_path)
                    logger.warning('a: {0}, b: {1}.'.format(a, b))

                if self.get_name_by_order:
                    channel_name = channel_no
                else:
                    channel_name = None

                channel = self.channel_data_class(current_channel_info, raw_data, self.duration(),
                                                  use_id_as_name=self.use_id_as_name, channel_name=channel_name)

                # Assign the channel either as normal channel or photodiode
                if channel.is_photodiode:
                    if channel.channel_name in photodiodes.keys():
                        # Check if current naming convention produces unique files
                        raise IOError('Trying to import two photodiodes with the same name')
                    photodiodes[channel.channel_name] = channel
                else:
                    if channel.channel_name in channels.keys():
                        # Check if current naming convention does not produce unique files
                        raise IOError('Trying to import two channels with the same name')
                    channels[channel.channel_name] = channel

        self.channels = channels
        self.photodiodes = photodiodes

        self._calculate_physical()

    def read_header(self, f):
        """ Read the header of an open Licel file.

        Parameters
        ----------
        f : file-like object
           An open file object.
        """
        # Read the first 3 lines of the header
        raw_info = {}
        channel_info = []

        # Read first line
        first_line = f.readline().decode().strip()
        raw_info['Filename'] = first_line
        self.header_lines.append(first_line)

        raw_info.update(self._read_second_header_line(f))

        raw_info.update(self._read_rest_of_header(f))

        # Update the object properties based on the raw info
        start_string = '%s %s' % (raw_info['start_date'], raw_info['start_time'])
        stop_string = '%s %s' % (raw_info['end_date'], raw_info['end_time'])
        date_format = '%d/%m/%Y %H:%M:%S'

        try:
            logger.debug('Creating timezone object %s' % self.licel_timezone)
            timezone = pytz.timezone(self.licel_timezone)
        except:
            raise ValueError("Cloud not create time zone object %s" % self.licel_timezone)

        # According to pytz docs, timezones do not work with default datetime constructor.
        local_start_time = timezone.localize(datetime.datetime.strptime(start_string, date_format))
        local_stop_time = timezone.localize(datetime.datetime.strptime(stop_string, date_format))

        # Only save UTC time.
        self.start_time = local_start_time.astimezone(pytz.utc)
        self.stop_time = local_stop_time.astimezone(pytz.utc)

        # Read the rest of the header.
        for c1 in range(int(raw_info['number_of_datasets'])):
            channel_line = f.readline().decode()
            channel_info.append(self.match_lines(channel_line, self.licel_file_channel_format))
            self.header_lines.append(channel_line.strip())

        self.raw_info = raw_info
        self.channel_info = channel_info

        self._assign_properties()

    def _assign_properties(self):
        """ Assign properties from the raw_info dictionary. """
        self.number_of_datasets = int(self.raw_info['number_of_datasets'])
        self.altitude = float(self.raw_info['altitude'])
        self.longitude = float(self.raw_info['longitude'])
        self.latitude = float(self.raw_info['latitude'])

        self.zenith_angle_raw = float(self.raw_info['zenith_angle'])

        logger.debug('Fix zenith angle? %s' % self.fix_zenith_angle)

        if self.fix_zenith_angle:
            logger.debug('Fixing zenith angle.')
            self.zenith_angle = self._correct_zenith_angle(self.zenith_angle_raw)
        else:
            self.zenith_angle = self.zenith_angle_raw

    @staticmethod
    def _correct_zenith_angle(zenith_angle):
        """ Correct zenith angle from Raymetrics convention (zenith = -90 degrees).

        Parameters
        ----------
        zenith_angle : float
           Zenith angle in Raymetrics convention.

        Returns
        -------
        corrected_angle : float
           Corrected zenith angle.
        """
        corrected_angle = 90 + zenith_angle
        return corrected_angle

    def _read_second_header_line(self, f):
        """ Read the second line of a licel file. """
        raw_info = {}

        second_line = f.readline().decode()
        self.header_lines.append(second_line.strip())
        # Many Licel files don't follow the licel standard. Specifically, the
        # measurement site is not always 8 characters, and can include white
        # spaces. For this, the site name is detect everything before the first
        # date. For efficiency, the first date is found by the first '/'.
        # e.g. assuming a string like 'Site name 01/01/2010 ...'

        site_name = second_line.split('/')[0][:-2]
        clean_site_name = site_name.strip()
        raw_info['site'] = clean_site_name
        self.site = clean_site_name

        raw_info.update(self.match_lines(second_line[len(clean_site_name) + 1:], self.licel_file_header_format[1]))
        return raw_info

    def _read_rest_of_header(self, f):
        """ Read the rest of the header lines, after line 2. """
        # Read third line
        third_line = f.readline().decode()
        self.header_lines.append(third_line.strip())

        raw_dict = self.match_lines(third_line, self.licel_file_header_format[2])
        return raw_dict

    def _calculate_physical(self):
        """ Calculate physical quantities from raw data for all channels in the file. """
        for channel in self.channels.values():
            channel.calculate_physical()

        for photodiode in self.photodiodes.values():
            photodiode.calculate_physical()

    def duration(self):
        """ Return the duration of the file.

        Returns
        -------
        : float
           The duration of the file in seconds.
        """
        dt = self.stop_time - self.start_time
        return dt.seconds

    def import_header_only(self):
        """ Import only the header lines, without reading the actual data."""
        with open(self.file_path, 'rb') as f:
            self.read_header(f)

    @property
    def has_photodiode(self):
        return len(self.photodiodes) != 0

    @staticmethod
    def match_lines(f1, f2):
        # TODO: Change this to regex?
        list1 = f1.split()
        list2 = f2.split()

        if len(list1) != len(list2):
            logging.debug("Channel parameter list has different length from LICEL specifications.")
            logging.debug("List 1: %s" % list1)
            logging.debug("List 2: %s" % list2)

        combined = list(zip(list2, list1))
        combined = dict(combined)
        return combined

    def save_as_csv(self, file_path=None, fill_value=-999):
        """ Save the Licel file in txt format.

        The format roughly follows the txt files created by Licel software. There are two main differences:
        a) Channel names are used as headers.
        b) Photon-counting data are given in shots, not in MHz.

        Parameters
        ----------
        file_path : str or None
           The output file path. If nan, the input file path is used with a .txt suffix.
        fill_value : float
           A fill value to be used in case of different length columns, e.g. when saving photodiode data.

        Returns
        -------
        file_path : str
           Returns the used file paths. This is useful when input file_path is None.
        """
        if file_path is None:
            file_path = self.file_path + ".csv"

        # Collect channel names and data
        column_names = []
        column_data = []

        for name, channel in self.channels.items():
            if channel.is_analog:
                column_name = "{0} (mV)".format(name)
            else:
                column_name = "{0} (counts)".format(name)

            column_names.append(column_name)
            column_data.append(channel.data)

        for name, photodiode in self.photodiodes.items():
            if 'PD' not in name:
                name = 'PD_' + name

            column_names.append(name)
            column_data.append(photodiode.data)

        column_data = self._common_length_array(column_data, fill_value)

        header_text = '\n'.join(self.header_lines) + '\n'
        column_header = ', '.join(column_names)

        np.savetxt(file_path, column_data.T,  fmt='%.4f', delimiter=',', header=header_text + column_header, comments='')

        return file_path

    @staticmethod
    def _common_length_array(array_list, fill_value):
        """ Make a signle array out of serveral 1D arrays with, possibly, different length"""

        lengths = [len(a) for a in array_list]

        if len(set(lengths)) == 1:
            output_array = np.array(array_list)
        else:
            dimensions = (len(lengths), max(lengths))
            output_array = np.ma.masked_all(dimensions)

            for n, array in enumerate(array_list):
                output_array[n, :len(array)] = array

            output_array.filled(fill_value)

        return output_array


class LicelChannel(LidarChannel):

    def __init__(self):
        self.name = None
        self.resolution = None
        self.points = None
        self.wavelength = None
        self.laser_used = None

        self.rc = []
        self.raw_info = []
        self.laser_shots = []
        self.duration = []
        self.discriminator = []
        self.hv = []
        self.data = {}

    def append_file(self, current_file, file_channel):
        """ Append file to the current object """

        self._assign_properties(current_file, file_channel)

        self.binwidth = self.resolution * 2. / c  # in seconds
        self.z = file_channel.z

        self.data[current_file.start_time] = file_channel.data
        self.raw_info.append(file_channel.raw_info)

        self.duration.append(file_channel.duration)
        self.laser_shots.append(file_channel.number_of_shots)
        self.discriminator.append(file_channel.discriminator)
        self.hv.append(file_channel.hv)

    @property
    def number_of_shots(self):
        """ Redundant, kept here for backward compatibility """
        return self.laser_shots

    def _assign_properties(self, current_file, file_channel):
        self._assign_unique_property('name', file_channel.channel_name)
        self._assign_unique_property('resolution', file_channel.dz)
        self._assign_unique_property('wavelength', file_channel.wavelength)
        self._assign_unique_property('points', file_channel.data_points)
        self._assign_unique_property('adcbits', file_channel.adcbits)
        self._assign_unique_property('active', file_channel.active)
        self._assign_unique_property('laser_used', file_channel.laser_used)
        self._assign_unique_property('analog_photon_string', file_channel.analog_photon_string)
        self._assign_unique_property('latitude', current_file.latitude)
        self._assign_unique_property('longitude', current_file.longitude)

    def _assign_unique_property(self, property_name, value):

        current_value = getattr(self, property_name, None)

        if current_value is None:
            setattr(self, property_name, value)
        else:
            if current_value != value:
                raise ValueError('Cannot combine channels with different values of {0}.'.format(property_name))

    @property
    def is_analog(self):
        return self.analog_photon_string == 'an'

    @property
    def is_photon_counting(self):
        return self.analog_photon_string == 'ph'

    def __unicode__(self):
        return "<Licel channel: %s>" % self.name

    def __str__(self):
        return str(self).encode('utf-8')


class PhotodiodeChannel(LicelChannel):

    def _assign_properties(self, current_channel, file_channel):
        """ In contrast with normal channels, don't check for constant points."""
        self._assign_unique_property('name', file_channel.channel_name)
        self._assign_unique_property('resolution', file_channel.dz)
        self._assign_unique_property('wavelength', file_channel.wavelength)
        self._assign_unique_property('adcbits', file_channel.adcbits)
        self._assign_unique_property('active', file_channel.active)
        self._assign_unique_property('laser_used', file_channel.laser_used)
        self._assign_unique_property('adcbits', file_channel.adcbits)
        self._assign_unique_property('analog_photon_string', file_channel.analog_photon_string)


class LicelLidarMeasurement(BaseLidarMeasurement):
    file_class = LicelFile
    channel_class = LicelChannel
    photodiode_class = PhotodiodeChannel

    def __init__(self, file_list=None, use_id_as_name=False, get_name_by_order=False, licel_timezone='UTC'):
        self.raw_info = {}  # Keep the raw info from the files
        self.durations = {}  # Keep the duration of the files
        self.laser_shots = []

        self.use_id_as_name = use_id_as_name
        self.get_name_by_order = get_name_by_order
        self.licel_timezone = licel_timezone
        self.photodiodes = collections.OrderedDict()

        super(LicelLidarMeasurement, self).__init__(file_list)

    def _import_file(self, filename):

        if filename in self.files:
            logger.warning("File has been imported already: %s" % filename)
        else:
            logger.debug('Importing file {0}'.format(filename))
            current_file = self.file_class(filename, use_id_as_name=self.use_id_as_name,
                                           get_name_by_order=self.get_name_by_order,
                                           licel_timezone=self.licel_timezone)
            self.raw_info[current_file.file_path] = current_file.raw_info
            self.durations[current_file.file_path] = current_file.duration()

            file_laser_shots = []

            self._create_or_append_channel(current_file)

            self.laser_shots.append(file_laser_shots)
            self.files.append(current_file.file_path)

    def _create_or_append_channel(self, current_file):

        for channel_name, channel in current_file.channels.items():
            if channel_name not in self.channels:
                self.channels[channel_name] = self.channel_class()
            self.channels[channel_name].append_file(current_file, channel)

        for photodiode_name, photodiode in current_file.photodiodes.items():
            if photodiode_name not in self.photodiodes:
                self.photodiodes[photodiode_name] = self.photodiode_class()
            self.photodiodes[photodiode_name].append_file(current_file, photodiode)

    def append(self, other):

        self.start_times.extend(other.start_times)
        self.stop_times.extend(other.stop_times)

        for channel_name, channel in self.channels.items():
            channel.append(other.channels[channel_name])

    def _get_duration(self, raw_start_in_seconds):
        """ Return the duration for a given time scale. If only a single
        file is imported, then this cannot be guessed from the time difference
        and the raw_info of the file are checked.
        """

        if len(raw_start_in_seconds) == 1:  # If only one file imported
            duration = next(iter(self.durations.values()))  # Get the first (and only) raw_info
            duration_sec = duration
        else:
            duration_sec = np.diff(raw_start_in_seconds)[0]

        return duration_sec

    def _get_custom_variables(self, channel_names):

        daq_ranges = np.ma.masked_all(len(channel_names))
        for n, channel_name in enumerate(channel_names):
            channel = self.channels[channel_name]
            if channel.is_analog:
                unique_values = list(set(channel.discriminator))
                if len(unique_values) > 1:
                    logger.warning(
                        'More than one discriminator levels for channel {0}: {1}'.format(channel_name, unique_values))
                daq_ranges[n] = unique_values[0]

        laser_shots = []
        for channel_name in channel_names:
            channel = self.channels[channel_name]
            laser_shots.append(channel.laser_shots)

        try:
            laser_shots = np.vstack(laser_shots).T
        except Exception as e:
            logger.error('Could not read laser shots as an array. Maybe files contain different number of channels?')
            raise e

        params = [{
            "name": "DAQ_Range",
            "dimensions": ('channels',),
            "type": 'd',
            "values": daq_ranges,
        }, {
            "name": "Laser_Shots",
            "dimensions": ('time', 'channels',),
            "type": 'i',
            "values": laser_shots,
        },
        ]

        return params

    def _get_custom_global_attributes(self):
        """
        NetCDF global attributes that should be included
        in the final NetCDF file.

        Currently the method assumes that all files in the measurement object have the same altitude, lat and lon
        properties.
        """
        logger.debug('Setting custom global attributes')
        logger.debug('raw_info keys: {0}'.format(self.raw_info.keys()))

        params = [{
            "name": "Altitude_meter_asl",
            "value": float(self.raw_info[self.files[0]]["altitude"])
        }, {
            "name": "Latitude_degrees_north",
            "value": float(self.raw_info[self.files[0]]["latitude"])
        }, {
            "name": "Longitude_degrees_east",
            "value": float(self.raw_info[self.files[0]]["longitude"])
        },
        ]

        return params

    def subset_by_channels(self, channel_subset):
        """
        Create a measurement object containing only a subset of  channels.

        This method overrides the parent method to add some licel-spefic parameters to the new object.

        Parameters
        ----------
        channel_subset : list
           A list of channel names (str) to be included in the new measurement object.

        Returns
        -------
        m : BaseLidarMeasurements object
           A new measurements object
        """
        new_measurement = super(LicelLidarMeasurement, self).subset_by_channels(channel_subset)

        new_measurement.raw_info = copy.deepcopy(self.raw_info)
        new_measurement.durations = copy.deepcopy(self.durations)
        new_measurement.laser_shots = copy.deepcopy(self.laser_shots)

        return new_measurement

    def subset_by_time(self, channel_subset):
        """
        Subsetting by time does not work yet with Licel files.

        This requires changes in generic.py
        """
        raise NotImplementedError("Subsetting by time, not yet implemented for Licel files.")

    def print_channels(self):
        """ Print the available channel information on the screen.
        """
        keys = sorted(self.channels.keys())

        print("Name  Wavelength  Mode  Resolution  Bins ")

        for key in keys:
            channel = self.channels[key]
            print("{0:<3}  {1:<10}  {2:<4}  {3:<10}  {4:<5}".format(channel.name, channel.wavelength,
                                                                    channel.analog_photon_string, channel.resolution,
                                                                    channel.points))


class LicelDivaLidarMeasurement(DivaConverterMixin, LicelLidarMeasurement):
    pass



import datetime as dt
from datetime import datetime
from pathlib import Path
from typing import Any
import numpy as np
import xarray as xr
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

from gfatpy.lidar.preprocessing import preprocess
from gfatpy.utils.plot import color_list


def binning(
    signal_an: np.ndarray[Any, np.dtype[np.float64]],
    signal_pc: np.ndarray[Any, np.dtype[np.float64]],
    pc_signal_channel: str,
    target_date: str | dt.date,
    pc_binning_range: tuple[float, float] = (0, 50),
    pc_bin_width: float = 1.0,    
    plot_dir: Path = Path.cwd(),
    savefig: bool = False,
):
    """Binning of AN and PC daily signals.

    Args:
        signal_an (np.ndarray): Analog signal.
        signal_pc (np.ndarray): Photon counting signal.
        pc_signal_channel (str): Channel name.
        target_date (str | datetime.date): Target date.
        pc_binning_range (tuple[float, float]): Range of PC signal. Default=(0, 50). In MHz.
        pc_bin_width (float): Width of the bin. Default=1.0. In MHz.        
        plot_dir (Path): Path to save the plot. Default=Path.cwd().
        savefig (bool): Save the plot. Default=False.
    """

    def plot_binning(
        signal_an: np.ndarray[Any, np.dtype[np.float64]],
        signal_pc: np.ndarray[Any, np.dtype[np.float64]],
        median_pc: np.ndarray[Any, np.dtype[np.float64]],
        median_an: np.ndarray[Any, np.dtype[np.float64]],
        std_an: np.ndarray[Any, np.dtype[np.float64]],
        pc_signal_channel: str,
        target_date: str,
        plot_dir: Path,
        savefig: bool,
    ):
        fig, ax = plt.subplots(figsize=[10, 5])
        ax.scatter(signal_pc, signal_an, c="grey", s=0.5, label="raw")
        ax.errorbar(
            median_pc,
            median_an,
            yerr=std_an,
            linestyle="None",
            marker=".",
            color="red",
            label="binning",
        )
        ax.set_xlabel("PC signal, [MHz]")
        ax.set_ylabel("dc-corrected AN signal, [mV]")
        ax.set_title(f"{pc_signal_channel} | {target_date}")
        ax.set_yscale("linear")
        ax.set_xlim(*pc_binning_range)
        ax.legend(fontsize=10, loc="upper right")
        if savefig:
            plt.savefig(plot_dir / f'binning_{pc_signal_channel}_{target_date}.png')
        plt.close(fig)

    if isinstance(target_date, str):        
        target_date = datetime.strptime(target_date, "%Y%m%d")

    # Apply binning to pc and an
    bins = np.arange(
        pc_binning_range[0], pc_binning_range[1] + pc_bin_width, pc_bin_width
    )

    # Use numpy.digitize to get the bin indices for each element in signal_pc
    bin_indices = np.digitize(signal_pc, bins)

    median_an = np.array([np.median(signal_an[bin_indices == i]) for i in range(1, len(bins))])

    std_an = np.array([np.std(signal_an[bin_indices == i]) for i in range(1, len(bins))])
    median_pc = np.array([np.median(signal_pc[bin_indices == i]) for i in range(1, len(bins))])
    isnan = np.logical_and(np.isnan(median_pc), np.isnan(median_an), np.isnan(std_an))

    # Plot binning
    plot_binning(
        signal_an,
        signal_pc,
        median_pc,
        median_an,
        std_an,
        pc_signal_channel,
        target_date.strftime("%Y%m%d"),
        plot_dir,
        savefig=savefig,
    )

    return median_pc[~isnan], median_an[~isnan], std_an[~isnan]


def get_valid_an_pc_values(
    signal_an: xr.DataArray,
    signal_pc: xr.DataArray,
    pc_threshold: float,
    an_threshold: float,
) -> tuple[np.ndarray[Any, np.dtype[np.float64]], np.ndarray[Any, np.dtype[np.float64]]]:
    """Filter analog and photoncouting signal using analog and photoncouting thresholds and removing NaNs.

    Args:
        signal_an (xr.DataArray): Analog signal.
        signal_pc (xr.DataArray): Photon counting signal.
        pc_threshold (float): Photon counting threshold in MHz.
        an_threshold (float): Analog threshold in mV.

    Returns:
        tuple[np.ndarray[Any, np.dtype[np.float64]], np.ndarray[Any, np.dtype[np.float64]]]: Filtered analog and photoncouting signal.
    """

    gl_condition = np.logical_and(
        signal_an.values > an_threshold, signal_pc.values < pc_threshold
    )
    an = np.concatenate(signal_an.where(gl_condition).values, axis=0)
    pc = np.concatenate(signal_pc.where(gl_condition).values, axis=0)
    isnan = np.logical_and(np.isnan(an), np.isnan(pc))
    an = an[~isnan]
    pc = pc[~isnan]
    return an, pc


def cost_function(
    tau_range: np.ndarray[Any, np.dtype[np.float64]],
    median_an: np.ndarray[Any, np.dtype[np.float64]],
    median_pc: np.ndarray[Any, np.dtype[np.float64]],
    std_an: np.ndarray[Any, np.dtype[np.float64]],
    pc_signal_channel: str,
    target_date: str | dt.date,
    savefig: bool,
    plot_dir: Path = Path.cwd(),
) -> tuple[np.ndarray[Any, np.dtype[np.float64]], float]:
    """Cost function to find the optimal dead time value (tau).

    Args:
        tau_range (np.ndarray): Candidate dead time values.
        median_an (np.ndarray): Binned valid analog signal from `.binning` function.
        median_pc (np.ndarray): Binned valid photon counting signal from `.binning` function.
        std_an (np.ndarray): Standard deviation of the binned analog signal from `.binning` function.
        pc_signal_channel (str): Channel name.
        target_date (str | datetime.date): Target date.
        savefig (bool): Save the plot. Default=False.
        plot_dir (Path): Path to save the plot. Default=Path.cwd().

    Returns:
        tuple[np.ndarray[Any, np.dtype[np.float64]], float]: Cost function and optimal dead time value.
    """
    
    def plot_cost_function(
        tau_range: np.ndarray[Any, np.dtype[np.float64]],
        J: np.ndarray[Any, np.dtype[np.float64]],
        pc_signal_channel: str,
        plot_dir: Path,
        target_date: str,
        optimal_tau_index: int,
        savefig: bool,
    ):
        logJ = np.log(J)
        optimal_tau = tau_range[optimal_tau_index]
        fig, ax = plt.subplots(figsize=[15, 5])
        ax.plot(tau_range, logJ, c="blue", lw=2, label="J", zorder=1)
        ax.scatter(
            optimal_tau,
            logJ[optimal_tau_index],
            marker="p",
            s=50,
            color="yellow",
            label=f"optimal tau = {optimal_tau:.2f} ns",
            zorder=2,
        )
        ax.set_ylabel("log(J), [#]")
        ax.set_xlabel(r"$\tau$, [ns]")
        ax.minorticks_on()
        ax.grid(True, which="both", linestyle=":", linewidth=0.5, color=[0.3, 0.3, 0.3])
        ax.set_title(f"{pc_signal_channel} | {target_date}")
        ax.legend(fontsize=10, loc="upper center")
        if savefig:
            plt.savefig(plot_dir / f'cost_function_{pc_signal_channel}_{target_date}.png')
        plt.close(fig)

    if isinstance(target_date, str):
        target_date = datetime.strptime(target_date, "%Y%m%d")

    J = np.nan * np.ones(len(tau_range))
    for idx, tau_ in enumerate(tau_range):
        dt_pc = median_pc / (1 - tau_ * median_pc * 1e-3)
        if not np.isnan(dt_pc).any() and not np.isnan(median_an).any():
            try:
                fit_values = np.polyfit(dt_pc, median_an, 1)
                virtual_an = np.polyval(fit_values, dt_pc)
            except Exception as e:                
                print(e)
                raise ValueError("Error in fitting binned an and pc")

            try:
                # Cost function
                idx_zero = std_an == 0
                J[idx] = np.sum(
                    (median_an[~idx_zero] - virtual_an[~idx_zero]) ** 2
                    / std_an[~idx_zero] ** 2
                ) / len(median_an)
            except Exception as e:                
                print(e)
                raise ValueError("Error in cost function")
    optimal_tau_index = np.log(J).argmin().astype(int)

    plot_cost_function(
        tau_range=tau_range,
        J=J,
        pc_signal_channel=pc_signal_channel,
        plot_dir=plot_dir,
        target_date=target_date.strftime("%Y%m%d"),
        optimal_tau_index=optimal_tau_index,
        savefig=savefig,
    )
    return J, tau_range[optimal_tau_index]


def plot_influence_of_dead_time_correction(
    tau_range: np.ndarray[Any, np.dtype[np.float64]],
    optimal_tau: float,
    median_an: np.ndarray[Any, np.dtype[np.float64]],
    median_pc: np.ndarray[Any, np.dtype[np.float64]],
    pc_signal_channel: str,
    target_date: str,
    plot_dir: Path,
    savefig: bool,
):
    fig, ax = plt.subplots(figsize=[15, 5])    
    h0_line, = ax.plot(median_pc, median_an, c="black", lw=1.5, zorder=2)
    colors = color_list(len(tau_range))
    handles = [None] * len(tau_range)
    for idx, tau_ in enumerate(tau_range):
        dt_pc = median_pc / (1 - tau_ * median_pc * 1e-3)
        handles[idx] = ax.scatter(dt_pc, median_an, #type: ignore
                                  color=colors[idx], alpha=0.7) 

    # Create a ScalarMappable object with the jet colormap and the range of values
    norm = mcolors.Normalize(vmin=tau_range.min(), vmax=tau_range.max())
    sm = cm.ScalarMappable(cmap='jet', norm=norm)
    sm.set_array([])  # Set a dummy array for the ScalarMappable

    # Add the colorbar to the figure
    cbar = plt.colorbar(sm, ax=ax, orientation='vertical', boundaries=tau_range[::100], cmap='jet')

    cbar.set_label(r'$\tau$, [ns]')

    opt_dt_pc = median_pc / (1 - optimal_tau * median_pc * 1e-3)
    h_opt, = ax.plot(opt_dt_pc,median_an, c="red", lw=2, zorder=2) #type: ignore
    ax.legend(
        [h0_line,h_opt], #type: ignore
        ['raw', f"optimal tau [{optimal_tau:.2f} ns]"],
        fontsize=10,
        loc="upper left")
    # ax.legend([h0, handles[0], handles[-1]], fontsize=10, loc="upper left")
    ax.set_xlabel("dt-corrected PC signal, [MHz]")
    ax.set_ylabel("dc-corrected AN signal, [mV]")
    ax.set_title(f"{target_date} | {pc_signal_channel}")
    ax.grid(True, which="both", linestyle=":", linewidth=0.5, color=[0.3, 0.3, 0.3])
    fig.tight_layout()
    if savefig:
        plt.savefig(plot_dir / f'pc_taus_{pc_signal_channel}_{target_date}.png')
    plt.close(fig)

def plot_gluing_thresholds(
    signal_an: xr.DataArray,
    signal_pc: xr.DataArray,
    pc_threshold: float,
    an_threshold: float,
    target_date: str,
    plot_dir: Path,
    savefig: bool,
    range_limits: tuple[float, float],
):
    pc_signal_channel = str(signal_pc.name).split("_")[-1]

    fig, ax = plt.subplots(figsize=[15, 5])
    lan = signal_an.plot.line(x="range", c="grey", lw=0.5, label="an", ax=ax) #type: ignore
    lpc = signal_pc.plot.line(x="range", c="black", lw=0.5, label="pc", ax=ax) #type: ignore

    l1 = plt.hlines(
        pc_threshold, xmin=0, xmax=10000, color="darkred", ls="--", label="Cmax"
    )
    l2 = plt.hlines(
        an_threshold,
        xmin=0,
        xmax=10000,
        color="violet",
        ls="--",
        label="bg_threshold_an",
    )

    ax.set_title(f"{pc_signal_channel} | {target_date}")
    ax.set_yscale("log")
    ax.set_ylim(0.01, 3000)
    ax.set_xlim(*range_limits)
    ax.legend(handles=[lan[0], lpc[0], l1, l2], fontsize=10, loc="upper right")

    if savefig:
        plt.savefig(plot_dir / f"gluing_thresholds_{pc_signal_channel}_{target_date}.png")
    plt.close(fig)


def dead_time_finder_by_channel(
    lidar: xr.Dataset,
    an_signal_channel: str,
    pc_signal_channel: str,
    pc_threshold: float,
    an_threshold: float,
    tau_range: list[float] | np.ndarray[Any, np.dtype[np.float64]] = np.arange(2, 10, 0.01),
    pc_binning_range: tuple[float, float] = (0., 50.),
    pc_bin_width: float = 1.0,
    plot_dir: Path = Path.cwd(),
    savefig: bool = False,
    range_limits: tuple[float, float] = (0, 10000),
) -> float:
    """Find the optimal dead time value for a given channel.

    Args:
        lidar (Path): Lidar data file.
        channel_an (str): Analog channel name.
        channel_pc (str): Photon counting channel name.
        pc_threshold (float): Photon counting threshold in MHz.
        an_threshold (float): Analog threshold in mV.
        tau_range (list): Candidate dead time values.
        pc_binning_range (tuple[float, float]): Range of PC signal. Default=(0, 50). In MHz.
        pc_bin_width (float): Width of the bin. Default=1.0. In MHz.
        plot_dir (Path): Path to save the plot. Default=Path.cwd().
        savefig (bool): Save the plot. Default=False.
        range_limits (tuple[float, float]): Range of the plot. Default=(0, 10000).

    Returns:
        float: Optimal dead time value.
    """

    target_date = lidar.time.values[0].astype(str).split("T")[0].replace("-", "")
    signal_an = lidar[f"signal_{an_signal_channel}"]
    signal_pc = lidar[f"signal_{pc_signal_channel}"]

    if isinstance(tau_range, list):
        tau_range = np.array(tau_range)

    plot_gluing_thresholds(
        signal_an=signal_an,
        signal_pc=signal_pc,
        pc_threshold=pc_threshold,
        an_threshold=an_threshold,
        target_date=target_date,
        plot_dir=plot_dir,
        savefig=savefig,
        range_limits=range_limits,
    )

    # get valid an and pc values
    an, pc = get_valid_an_pc_values(
        signal_an=signal_an,
        signal_pc=signal_pc,
        pc_threshold=pc_threshold,
        an_threshold=an_threshold,
    )

    if len(an) < 100 or len(pc) < 100:
        print(f"Empty arrays for {pc_signal_channel} | {target_date}")
        return np.nan

    # apply binning
    median_pc, median_an, std_an = binning(
        signal_an=an,
        signal_pc=pc,
        pc_signal_channel=pc_signal_channel,
        target_date=target_date,
        pc_binning_range=pc_binning_range,
        pc_bin_width=pc_bin_width,
        plot_dir=plot_dir,
        savefig=savefig,
    )
    

    # Retrieve cost function
    J, optimal_tau = cost_function(
        tau_range=tau_range,
        median_an=median_an,
        median_pc=median_pc,
        std_an=std_an,
        pc_signal_channel=pc_signal_channel,
        target_date=target_date,
        savefig=savefig,
        plot_dir=plot_dir,
    )    

    # Plot influence of dead time correction
    plot_influence_of_dead_time_correction(
        tau_range=tau_range,
        optimal_tau=optimal_tau,
        median_an=median_an,
        median_pc=median_pc,
        pc_signal_channel=pc_signal_channel,
        target_date=target_date,
        plot_dir=plot_dir,
        savefig=savefig,
    )

    #Check if optimal tau is out of range
    if optimal_tau == tau_range[0] or optimal_tau == tau_range[-1]:
        print(f'Warning! {datetime} | {pc_signal_channel} : tau out of range')
        return np.nan
    print(f"{datetime} | Optimal tau for {pc_signal_channel} = {optimal_tau} ns")
    
    return optimal_tau


def estimate_daily_dead_time(
    file: Path | str,
    target_pc_channels: list[str] | None = None,
    tau_dir: Path | str = Path.cwd(),
    tau_range: list | np.ndarray[Any, np.dtype[np.float64]] = np.arange(2, 10, 0.01),
    crop_range: tuple = (0, 15000),
    pc_threshold: float = 50,
    an_threshold: float = 0.1,
    pc_binning_range: tuple = (0, 50),
    pc_bin_width: float = 1.,
    range_limits: tuple[float, float] = (0, 10000),
    savefig: bool = False,
    plot_dir: Path = Path.cwd(),
) -> Path:
    """Estimate the optimal dead time for a given list of channels of an specific daily lidar file and save the results in a netcdf file.

    Args:
        file (Path | str): Lidar data file.
        target_pc_channels (list[str]): List of target photon counting channels. Default=None (all channels).
        tau_dir (Path | str | None, optional): Directory to save the netcdf with the optimal dead time values. Defaults to None.
        tau_range (list | np.ndarray[Any, np.dtype[np.float64]], optional): Candidate dead time values. Defaults to np.arange(2, 10, 0.01).
        crop_range (tuple, optional): Range to crop the lidar data. Defaults to (0, 15000).
        pc_threshold (float, optional): Photon counting threshold in MHz. Defaults to 50.
        an_threshold (float, optional): Analog threshold in mV. Defaults to 0.1.
        pc_binning_range (tuple, optional): Range of PC signal. Defaults to (0, 50).
        pc_bin_width (float, optional): Width of the bin. Defaults to 1.0.
        range_limits (tuple[float, float], optional): Range of the plot. Defaults to (0, 10000).
        savefig (bool, optional): Save the plot. Defaults to False.
        plot_dir (Path, optional): Path to save the plot. Defaults to Path.cwd().

    Raises:
        FileNotFoundError: File does not exist.
        TypeError: File is not a netcdf file.
        ValueError: tau_range must be a 1D array.
        ValueError: Error creating {plot_dir}

    Returns:
        Path: Path to the netcdf file with the optimal dead time values.
    """

    # Management file type
    if isinstance(file, str):
        file = Path(file)
    if file.exists() == False:
        raise FileNotFoundError(f"{file} does not exist")
    if file.suffix != ".nc":
        raise TypeError(f"{file} is not a netcdf file")

    # Management tau_dir type
    if isinstance(tau_dir, str):
        tau_dir = Path(tau_dir)
    if tau_dir.exists() == False:
        try:
            tau_dir.mkdir(parents=True, exist_ok=True)
        except Exception as e:
            print(e)
            raise ValueError(f"Error creating {tau_dir}")

    # Management tau_range type
    if isinstance(tau_range, list):
        tau_range = np.array(tau_range)
    if tau_range.ndim != 1:
        raise ValueError(f"tau_range must be a 1D array")

    # Read
    lidar = preprocess(
        file,
        apply_dt=False,
        save_bg=False,
        save_dc=True,
        apply_bz=True,
        crop_ranges=crop_range,
    )
    target_date = lidar.time.values[1].astype(str).split("T")[0].replace("-", "")
    
    if savefig:
        plot_dir = tau_dir / f"plots_{target_date}"
        if plot_dir.exists() == False:
            try:
                plot_dir.mkdir(parents=True, exist_ok=True)
            except Exception as e:
                print(e)
                raise ValueError(f"Error creating {plot_dir}")

    # Create tau dictionary
    optimal_taus = {}

    for channel_ in [
        channel_ for channel_ in lidar.channel.values if channel_.endswith("a")
    ]:
        an_signal_channel, pc_signal_channel = channel_, channel_.replace("a", "p")
            
        if pc_signal_channel in lidar.channel:

            if target_pc_channels is not None:                    
                if pc_signal_channel not in target_pc_channels:
                    continue
                           
            optimal_taus[pc_signal_channel] = dead_time_finder_by_channel(
                lidar=lidar,
                an_signal_channel=an_signal_channel,
                pc_signal_channel=pc_signal_channel,
                pc_threshold=pc_threshold,
                an_threshold=an_threshold,
                pc_binning_range=pc_binning_range,
                pc_bin_width=pc_bin_width,
                plot_dir=plot_dir,
                savefig=savefig,
                tau_range=tau_range,
                range_limits=range_limits,
            )

    # Save optimal_taus to a xarray.Dataset 
    tau_dataset = xr.Dataset( { "dead_time": xr.DataArray( list(optimal_taus.values()), coords=[list(optimal_taus.keys())], dims=["channel"], ) } )
    tau_dataset["dead_time"].attrs["units"] = "ns"
    tau_dataset["dead_time"].attrs["long_name"] = "Dead time"
    tau_dataset.attrs["date"] = target_date
    output_file = tau_dir / f"alh_dead-time_{target_date}_2.nc"
    tau_dataset.to_netcdf(output_file)
    return output_file



import numpy as np
import xarray as xr
from scipy.constants import Boltzmann

from gfatpy.atmo import atmo, ecmwf
from gfatpy.atmo.rayleigh import molecular_properties
from gfatpy.lidar.utils.types import ParamsDict
from gfatpy.lidar.utils.utils import extrapolate_beta_with_angstrom
from gfatpy.lidar.utils.utils import sigmoid


def generate_particle_properties(
    ranges: np.ndarray,
    wavelength: float,
    ae: float | tuple[float, float] = (1.5, 0),
    lr: float | tuple[float, float] = (75, 45),
    synthetic_beta: float | tuple[float, float] = (2.5e-6, 2.0e-6),
    sigmoid_edge: float | tuple[float, float] = (2500, 5000),
) -> np.ndarray:
    """_summary_

    Args:
        ranges (np.ndarray): ranges
        wavelength (float): wavelength
        fine_ae (float): fine-mode Angstrom exponent
        coarse_ae (float): coarse-mode Angstrom exponent
        fine_beta532 (float, optional): fine-mode backscatter coefficient at 532 nm. Defaults to 2.5e-6.
        coarse_beta532 (float, optional): coarse-mode backscatter coefficient at 532 nm. Defaults to 2.0e-6.

    Returns:
        np.ndarray: particle backscatter coefficient profile
    """
    if isinstance(ae, tuple):
        fine_ae = ae[0]
        coarse_ae = ae[1]
    else:
        fine_ae = ae
        coarse_ae = ae

    if isinstance(lr, tuple):
        fine_lr = lr[0]
        coarse_lr = lr[1]
    else:
        fine_lr = lr
        coarse_lr = lr

    if isinstance(sigmoid_edge, tuple):
        sigmoid_edge_fine = sigmoid_edge[0]
        sigmoid_edge_coarse = sigmoid_edge[1]
    else:
        sigmoid_edge_fine = sigmoid_edge
        sigmoid_edge_coarse = sigmoid_edge

    if isinstance(synthetic_beta, tuple):
        fine_beta532 = synthetic_beta[0]
        coarse_beta532 = synthetic_beta[1]
    else:
        fine_beta532 = synthetic_beta
        coarse_beta532 = synthetic_beta

    beta_part_fine_532 = sigmoid(
        ranges, sigmoid_edge_fine, 1 / 60, coeff=-fine_beta532, offset=fine_beta532
    )
    beta_part_coarse_532 = sigmoid(
        ranges,
        sigmoid_edge_coarse,
        1 / 60,
        coeff=-coarse_beta532,
        offset=coarse_beta532,
    )

    beta_part_fine = extrapolate_beta_with_angstrom(
        beta_part_fine_532, 532, wavelength, fine_ae
    )

    beta_part_coarse = extrapolate_beta_with_angstrom(
        beta_part_coarse_532, 532, wavelength, coarse_ae
    )

    beta_total = beta_part_fine + beta_part_coarse

    alpha_part_fine = fine_lr * beta_part_fine
    alpha_part_coarse = coarse_lr * beta_part_coarse

    alpha_total = alpha_part_fine + alpha_part_coarse

    return (
        beta_part_fine,
        beta_part_coarse,
        beta_total,
        alpha_part_fine,
        alpha_part_coarse,
        alpha_total,
    ) # type: ignore


def synthetic_signals(
    ranges: np.ndarray,
    wavelengths: float | tuple[float, float] = 532,
    overlap_midpoint: float = 800,
    k_lidar: float | tuple[float, float] = (1e10, 1e9),
    ae: float | tuple[float, float] = (1.5, 0),
    lr: float | tuple[float, float] = (75, 45),
    synthetic_beta: float | tuple[float, float] = (2.5e-6, 2.0e-6),
    sigmoid_edge: float | tuple[float, float] = (2500, 5000),
    force_zero_aer_after_bin: int | None = None,
    paralell_perpendicular_ratio: float = 0.33,
    meteo_profiles: tuple[np.ndarray, np.ndarray] | None = None,
    apply_overlap: bool = True,
) -> tuple[np.ndarray, np.ndarray | None, ParamsDict]:
    """It generates synthetic lidar signal.

    Args:
        ranges (np.ndarray): Range
        wavelength (float, optional): Wavelength. Defaults to 532.
        overlap_midpoint (float, optional): _description_. Defaults to 600.
        k_lidar (float, optional): Lidar constant calibration. Defaults to 4e9.
        wavelength_raman (float | None, optional): Raman wavelength. Defaults to None. If None, signal is elastic.
        paralell_perpendicular_ratio (float, optional): _description_. Defaults to 0.33.
        particle_lratio (float, optional): _description_. Defaults to 45.
        force_zero_aer_after_bin (int | None, optional): _description_. Defaults to None.

    Returns:
        tuple[np.ndarray, ParamsDict]: _description_
    """

    z = ranges.astype(np.float64)

    # Overlap
    if apply_overlap:
        overlap = sigmoid(
            z.astype(np.float64),
            overlap_midpoint,
            1 / 150.,
            offset=0.,
        )

        # overlap[overlap < 9e-3] = 0
        # overlap[overlap > 0.999] = 1
    else:
        overlap = np.ones_like(z)

    if isinstance(lr, float):
        lr = (lr, lr)
    if isinstance(ae, float):
        ae = (ae, ae)
    if isinstance(synthetic_beta, float):
        synthetic_beta = (synthetic_beta, synthetic_beta)

    if isinstance(k_lidar, float):
        k_lidar_elastic = k_lidar
    else:
        k_lidar_elastic, k_lidar_raman = k_lidar

    # Check temperature and pressure profiles
    if meteo_profiles is None:
        # ecmwf_data = ecmwf.get_ecmwf_temperature_pressure(datetime(2022, 9, 3), heights=z)
        # P = ecmwf_data.pressure.values
        # T = ecmwf_data.temperature.values
        P, T, _ = atmo.standard_atmosphere(z)
    else:
        # check length of meteo_profiles with z
        if len(meteo_profiles[0]) != len(z):
            raise ValueError("Length of meteo_profiles must be equal to length of z")
        else:
            P = meteo_profiles[0]
            T = meteo_profiles[1]

    # Check if wavelength is a tuple
    if isinstance(wavelengths, tuple):
        wavelength = wavelengths[0]
        wavelength_raman = wavelengths[1]
    else:
        wavelength = wavelengths
        wavelength_raman = None

    # Generate molecular profiles for elastic wavelength
    mol_properties = molecular_properties(532, P, T, heights=z)
    
    # TODO
    # Particle elastic
    # 0.33 para polvo desértico
    # 0.0034 para parte molecular
    # Calcular beta_part parallel = total*(1-0.33)
    # Calcular beta_part perpendicular = total*(0.33)
    # Análogo con otro coeficiente para la parte molecular

    # beta_part_fine, beta_part_coarse, beta_total, alpha_part_fine, alpha_part_coarse, alpha_total
    (
        _,
        _,
        beta_part,
        alpha_part_fine,
        alpha_part_coarse,
        alpha_part,
    ) = generate_particle_properties(
        ranges, wavelength, ae=ae, lr=lr, synthetic_beta=synthetic_beta, sigmoid_edge=sigmoid_edge
    )

    # Elastic transmittance
    # T_elastic = np.exp(-cumulative_trapezoid(alpha_mol+ alpha_part, z, initial=0))  # type: ignore
    T_elastic = atmo.transmittance(mol_properties["molecular_alpha"] + alpha_part, z)
    # Elastic signal
    P_elastic = (
        k_lidar_elastic
        * (overlap / z**2)
        * (mol_properties["molecular_beta"] + beta_part)
        * T_elastic**2
    )

    clean_attenuated_molecular_beta = (
        mol_properties["molecular_beta"]
        * atmo.transmittance(mol_properties["molecular_alpha"], z) ** 2
    )

    # Save parameters to create synthetic elastic signal
    params: ParamsDict = {
        "particle_beta": beta_part,
        "particle_alpha": alpha_part,
        "molecular_beta": mol_properties["molecular_beta"],
        "molecular_alpha": mol_properties["molecular_alpha"],
        "lidar_ratio": lr,
        "attenuated_molecular_backscatter": clean_attenuated_molecular_beta,
        "transmittance_elastic": T_elastic,
        "overlap": overlap,
        "k_lidar": k_lidar,
        "particle_angstrom_exponent": ae,
        "synthetic_beta": synthetic_beta,
        "temperature": T,
        "pressure": P,
    } # type: ignore

    # Raman signal
    if wavelength_raman is not None:
        # Generate molecular profiles for raman wavelength
        mol_properties_raman = molecular_properties(wavelength_raman, P, T, heights=z)

        # Alpha particle raman
        alpha_part_fine_raman = alpha_part_fine * (wavelength_raman / wavelength) ** (
            -ae[0]
        )
        alpha_part_coarse_raman = alpha_part_coarse * (
            wavelength_raman / wavelength
        ) ** (-ae[1])
        alpha_part_raman = alpha_part_fine_raman + alpha_part_coarse_raman

        # Transmittance Raman
        # T_raman = np.exp(-cumulative_trapezoid(alpha_mol_raman + alpha_part_raman, z, initial=0))  # type: ignore
        T_raman = atmo.transmittance(
            mol_properties_raman["molecular_alpha"] + alpha_part_raman, z
        )
        P_raman = (
            k_lidar_raman
            * (overlap / z**2)
            * mol_properties_raman["molecular_beta"]
            * T_elastic
            * T_raman
        )

        clean_attenuated_molecular_beta_raman: xr.DataArray = (
            mol_properties_raman["molecular_beta"]
            * atmo.transmittance(mol_properties_raman["molecular_alpha"], z)
            * atmo.transmittance(mol_properties["molecular_alpha"], z)
        ) # type: ignore

        params["molecular_alpha_raman"] = mol_properties_raman["molecular_alpha"] # type: ignore
        params["molecular_beta_raman"] = mol_properties_raman["molecular_beta"] # type: ignore
        params["attenuated_molecular_backscatter_raman"] = (
            clean_attenuated_molecular_beta_raman, 
        ) # type: ignore
        params["transmittance_raman"] = T_raman
        params["overlap"] = overlap

    else:
        P_raman = None

    if force_zero_aer_after_bin is not None:
        alpha_part[force_zero_aer_after_bin:] = 0
        beta_part[force_zero_aer_after_bin:] = 0
    return P_elastic, P_raman, params



import datetime as dt
from pathlib import Path
from pdb import set_trace
from typing import Tuple
import numpy as np
import pandas as pd
import xarray as xr
import matplotlib.pyplot as plt
from scipy.signal import savgol_filter

from gfatpy.atmo.atmo import transmittance
from gfatpy.utils import utils
from gfatpy.atmo.rayleigh import molecular_properties
from gfatpy.lidar.utils.file_manager import channel2info
from gfatpy.lidar.utils.utils import LIDAR_INFO, signal_to_rcs
# from gfatpy.utils.durbin_watson import linear_fit
from gfatpy.utils.optimized import moving_linear_fit


def _get_window_size(
    reference_ranges: tuple[float, float], ranges: np.ndarray[float]
) -> int:
    """It provides the window size of a given reference range (e.g., 6000 - 7000 meters).

    Args:
        reference_ranges (tuple[float, float]): Reference range in meters.
        ranges (np.dnarray[float]): Range array in meters.

    Returns:
        int: Window size.
    """
    dz = np.median(np.diff(ranges))  # type: ignore
    reference_idxs = (
        np.floor(reference_ranges[0] / dz).astype(int),
        np.floor(reference_ranges[1] / dz).astype(int),
    )
    return reference_idxs[1] - reference_idxs[0]


def _attenuated_backscatter(
    rcs: xr.DataArray,
    attenuated_molecular_backscatter: xr.DataArray,
    reference_range: Tuple[float, float],
) -> xr.DataArray:
    """Calculate the attenuated backscatter.

    Args:

        - channel (str): Channel to calculate the Rayleigh Fit.
        - rcs (xr.DataArray): Range Corrected Signal.
        - attenuated_molecular_backscatter (xr.DataArray): Attenuated Molecular Backscatter.
        - reference_range (Tuple[float, float]): Reference range to normalize.

    Returns:

        - xr.DataArray: Attenuated backscatter.
    """

    attenuated_backscatter = attenuated_molecular_backscatter.sel(
        range=slice(*reference_range)
    ).mean("range") * (rcs / rcs.sel(range=slice(*reference_range)).mean("range"))

    return attenuated_backscatter


def _residual(
    attenuated_backscatter: xr.DataArray, attenuated_molecular_backscatter: xr.DataArray
) -> xr.DataArray:
    """Calculate the residual between the attenuated backscatter and the attenuated molecular backscatter.

    Args:
        attenuated_backscatter (xr.DataArray): Attenuated backscatter.
        attenuated_molecular_backscatter (xr.DataArray): Attenuated molecular backscatter.

    Returns:
        xr.DataArray: Residual (difference) between the attenuated backscatter and the attenuated molecular backscatter.
    """

    residual = attenuated_backscatter - attenuated_molecular_backscatter
    residual.attrs = {
        "long_name": "residual between attenuated backscatter and attenuated molecular backscatter",
        "units": "m{-1} sr{-1}",
    }
    return residual


def _attenuated_backscattering_ratio(
    attenuated_backscatter: xr.DataArray,
    attenuated_molecular_backscatter: xr.DataArray,
    reference_ranges: tuple[float, float],
) -> xr.DataArray:
    """Calculate the attenuated backscattering ratio.

    Args:
        attenuated_backscatter (xr.DataArray): Attenuated backscatter.
        attenuated_molecular_backscatter (xr.DataArray): Attenuated molecular backscatter.
        reference_ranges (tuple[float, float]): Reference range to normalize.

    Returns:
        xr.DataArray: Attenuated backscattering ratio.
    """
    window_size_bin = _get_window_size(
        reference_ranges=reference_ranges, ranges=attenuated_backscatter.range.values
    )

    attenuated_backscattering_ratio = (
        (attenuated_backscatter / attenuated_molecular_backscatter)
        .rolling(range=window_size_bin, center=True)
        .mean("range")
    )
    attenuated_backscattering_ratio.attrs = {
        "long_name": "attenuated backscattering ratio",
        "units": "#",
    }

    return attenuated_backscattering_ratio


def _extinction_from_attenuated_backscatter(
    attenuated_backscatter: xr.DataArray,
    attenuated_molecular_backscatter: xr.DataArray,
    reference_ranges: tuple[float, float],
) -> Tuple[xr.DataArray, xr.DataArray, xr.DataArray]:
    """Calculate the extinction from the attenuated backscatter. The Durbin-Watson statistic is also calculated.

    Args:

        - attenuated_backscatter (xr.DataArray): Attenuated backscatter.
        - reference_ranges (tuple[float, float]): Aerosol-free region reference range.

    Returns:

        - Tuple[xr.DataArray, xr.DataArray]: Extinction, its standard deviation and Durbin-Watson statistic.

    References:

        - Baars, H., et al. (2016). https://doi.org/10.5194/acp-16-5111-2016
    """

    ranges = attenuated_backscatter.range.values
    window_size_bin = _get_window_size(reference_ranges=reference_ranges, ranges=ranges)

    results = moving_linear_fit(
        ranges,
        attenuated_backscatter.values/attenuated_molecular_backscatter.values,
        **{"length": window_size_bin, "degree": 1},
    )
    extinction = -0.5*results["slope"]
    std_extinction = 0.5*results["std_slope"]

    extinction = xr.DataArray(
        data=extinction,
        dims=["range"],
        coords={"range": ranges},
        attrs={
            "long_name": "extinction",
            "units": "m{-1} sr{-1}",
        },
    )
    std_extinction = xr.DataArray(
        data=std_extinction,
        dims=["range"],
        coords={"range": ranges},
        attrs={
            "long_name": "std_extinction",
            "units": "m{-1} sr{-1}",
        },
    )

    durbin_watson_statistic = xr.DataArray(
        data=d,
        dims=["range"],
        coords={"range": ranges},
        attrs={
            "long_name": "Durbin-Watson statistic",
            "units": "#",
        },
    )

    return extinction, std_extinction, durbin_watson_statistic


def _slope_attenuated_backscatter(
    attenuated_backscatter: xr.DataArray,
    reference_ranges: tuple[float, float],
) -> xr.DataArray:
    """Calculate the slope of the attenuated backscatter.

    Args:
        attenuated_backscatter (xr.DataArray): Attenuated backscatter.
        reference_ranges (tuple[float, float]): Reference range to normalize.

    Returns:
        xr.DataArray: Slope of the attenuated backscatter.
    """

    ranges = attenuated_backscatter.range.values
    dz = np.median(np.diff(ranges))  # type: ignore
    window_size_bin = _get_window_size(reference_ranges=reference_ranges, ranges=ranges)

    # Apply derivative
    slope_attenuated_backscatter = savgol_filter(
        1e6 * attenuated_backscatter.values,
        window_size_bin,
        1,
        deriv=1,
        delta=dz,
        mode="nearest",
        cval=np.nan,
    )  # Calculate 1st derivative

    slope_attenuated_backscatter = xr.DataArray(
        slope_attenuated_backscatter,
        dims=["range"],
        coords={"range": ranges},
        attrs={"long_name": "slope of attenuated backscatter", "units": "m{-2} sr{-1}"},
    )
    return slope_attenuated_backscatter


def _slope_attenuated_molecular_backscatter(
    attenuated_molecular_backscatter: xr.DataArray,
    reference_ranges: tuple[float, float],
) -> xr.DataArray:
    """Calculate the slope of the attenuated molecular backscatter.

    Args:
        attenuated_molecular_backscatter (xr.DataArray): Attenuated molecular backscatter.
        reference_ranges (tuple[float, float]): Reference range to normalize.

    Returns:
        xr.DataArray: Slope of the attenuated molecular backscatter.
    """

    ranges = attenuated_molecular_backscatter.range.values
    dz = np.median(np.diff(ranges))  # type: ignore
    window_size_bin = _get_window_size(reference_ranges=reference_ranges, ranges=ranges)

    # Apply derivative
    slope_attenuated_molecular_backscatter = savgol_filter(
        1e6 * attenuated_molecular_backscatter.values,
        window_size_bin,
        1,
        deriv=1,
        delta=dz,
        mode="nearest",
        cval=np.nan,
    )  # Calculate 1st derivative

    slope_attenuated_molecular_backscatter = xr.DataArray(
        slope_attenuated_molecular_backscatter,
        dims=["range"],
        coords={"range": ranges},
        attrs={
            "long_name": "slope of attenuated molecular backscatter",
            "units": "m{-2} sr{-1}",
        },
    )
    return slope_attenuated_molecular_backscatter


def _snr(
    signal: xr.DataArray,
    reference_ranges: tuple[float, float],
    mode: str = "analog",
) -> xr.DataArray:
    """Calculate the signal-to-noise ratio (SNR) of the analog signal.

    Args:
        signal (xr.DataArray): Signal.
        noise (xr.DataArray): Noise.
        reference_ranges (tuple[float, float]): Reference range to normalize.

    Returns:
        xr.DataArray: Signal-to-noise ratio (SNR).
    """

    ranges = signal.range.values
    window_size_bin = _get_window_size(reference_ranges=reference_ranges, ranges=ranges)

    if mode == "analog":
        signal_ = signal.rolling(range=window_size_bin, center=True).mean("range")
        noise_ = signal.rolling(range=window_size_bin, center=True).std("range")
    else:
        # TODO: implement equation A10 from the paper Baars et al. (2016) DOI: 10.5194/acp-16-5111-2016
        signal_ = signal.rolling(range=window_size_bin, center=True).mean("range")
        noise_ = signal.rolling(range=window_size_bin, center=True).std("range")
        raise Warning(
            "Only analog mode is implemented. Retrieval is made with analog equations."
        )

    snr = signal_ / noise_
    snr.attrs = {
        "long_name": "signal-to-noise ratio",
        "units": "#",
    }
    return snr


def get_mask_attenuated_backscattering_ratio(
    attenuated_backscattering_ratio: xr.DataArray,
    relative_difference_threshold: float = 0.05,
) -> xr.DataArray:
    """Get mask for attenuated backscattering ratio cloase to one with relative difference lower than a threshold.

    Args:
        attenuated_backscattering_ratio (xr.DataArray): Attenuated backscattering ratio.
        relative_difference_threshold (float, optional): Relative difference threshold. Defaults to 0.05.

    Returns:
        xr.DataArray: Mask for attenuated backscattering ratio.
    """
    # Get boolean variable where the attenuated backscattering ratio is within the thresholds
    mask = np.isclose(
        attenuated_backscattering_ratio, 1, rtol=relative_difference_threshold
    )

    mask = xr.DataArray(
        mask, dims=["range"], coords={"range": attenuated_backscattering_ratio.range}
    )
    mask.attrs = {
        "long_name": "mask for attenuated backscattering ratio within thresholds",
        "thresholds": f"relative difference lower than {relative_difference_threshold}",
        "units": "#",
    }
    return mask


def get_mask_extinction(
    extinction: xr.DataArray,
    std_extinction: xr.DataArray,
) -> xr.DataArray:
    """Get mask for extinction larger than the standard deviation.

    Args:
        extinction (xr.DataArray): Extinction.
        std_extinction (xr.DataArray): Standard deviation of the extinction.

    Returns:
        xr.DataArray: Mask for extinction > extinction std.
    """
    # Get boolean variable where the extinction is within the standard deviation
    mask = extinction > std_extinction
    mask = xr.DataArray(mask, dims=["range"], coords={"range": extinction.range})
    mask.attrs = {
        "long_name": "mask for extinction larger than standard deviation",
        "units": "#",
    }
    return mask


def get_mask_durbin_watson(
    durbin_watson_statistic: xr.DataArray,
    threshold_limits: tuple[float, float] = (1, 3),
) -> xr.DataArray:
    """Get mask for Durbin-Watson statistic within threshold limits.

    Args:
        durbin_watson_statistic (xr.DataArray): Durbin-Watson statistic.
        threshold_limits (tuple[float, float], optional): Threshold limits. Defaults to (1, 3).

    Returns:
        xr.DataArray: Mask for Durbin-Watson statistic.
    """

    # Get boolean variable where the Durbin-Watson statistic is within the threshold limits
    mask = (durbin_watson_statistic > threshold_limits[0]) & (
        durbin_watson_statistic < threshold_limits[1]
    )
    mask = xr.DataArray(
        mask, dims=["range"], coords={"range": durbin_watson_statistic.range}
    )
    mask.attrs = {
        "long_name": "mask for Durbin-Watson statistic within threshold limits",
        "thresholds": f"{threshold_limits[0]} < D-W < {threshold_limits[1]}",
        "units": "#",
    }
    return mask


def get_mask_nsr(
    nsr: xr.DataArray, threshold_limits: tuple[float, float] = (0.0, 0.15)
) -> xr.DataArray:
    """Get mask for NSR within threshold limits.

    Args:
        nsr (xr.DataArray): Noise-to-signal ratio.
        threshold_limits (tuple[float, float], optional): Threshold limits. Defaults to (0.0, 0.15).

    Returns:
        xr.DataArray: Mask for NSR.
    """
    # Get boolean variable where the Durbin-Watson statistic is within the threshold limits
    mask = (nsr > threshold_limits[0]) & (nsr < threshold_limits[1])
    mask = xr.DataArray(mask, dims=["range"], coords={"range": nsr.range})
    mask.attrs = {
        "long_name": "mask for NSR within threshold limits",
        "thresholds": f"{threshold_limits[0]} < NSR < {threshold_limits[1]}",
        "units": "#",
    }
    return mask


def get_mask_residual(
    attenuated_backscatter: xr.DataArray,
    attenuated_molecular_backscatter: xr.DataArray,
    reference_ranges: tuple[float, float],
) -> xr.DataArray:
    """Get mask for residual within the standard deviation.

    Args:
        attenuated_backscatter (xr.DataArray): Attenuated backscatter.
        attenuated_molecular_backscatter (xr.DataArray): Attenuated molecular backscatter.
        reference_ranges (tuple[float, float]): Aerosol-free reference range.

    Returns:
        xr.DataArray: Mask for residual.
    """
    window_size_bin = _get_window_size(
        reference_ranges, attenuated_backscatter.range.values
    )

    # Get boolean variable where the residual is within the standard deviation
    att_beta_mean = attenuated_backscatter.rolling(
        range=window_size_bin, center=True
    ).mean("range")
    att_beta_mean = att_beta_mean.where(~np.isnan(att_beta_mean), drop=True)
    att_beta_std = attenuated_backscatter.rolling(
        range=window_size_bin, center=True
    ).std("range")
    att_beta_std = att_beta_std.where(~np.isnan(att_beta_std), drop=True)
    att_mol_beta_mean = attenuated_molecular_backscatter.rolling(
        range=window_size_bin, center=True
    ).mean("range")
    att_mol_beta_mean = att_mol_beta_mean.where(~np.isnan(att_mol_beta_mean), drop=True)
    att_mol_beta_std = attenuated_molecular_backscatter.rolling(
        range=window_size_bin, center=True
    ).std("range")    
    att_mol_beta_std = att_mol_beta_std.where(~np.isnan(att_mol_beta_std), drop=True)

    mask = (att_beta_mean + att_beta_std) > (att_mol_beta_mean - att_mol_beta_std)
    # mask = xr.DataArray(
    #     mask, dims=["range"], coords={"range": attenuated_backscatter.range}
    # )
    mask.attrs = {
        "long_name": "mask for residual within standard deviation",
        "units": "#",
    }
    
    # breakpoint()
    # fig, ax = plt.subplots()
    # att_mol_beta_mean.plot(ax=ax, ls = '--', label='Attenuated Molecular Backscatter')    
    # att_beta_mean.plot(ax=ax, ls = '--', label='Attenuated Backscatter')
    # (att_mol_beta_mean - att_mol_beta_std).plot(ax=ax, label='Attenuated Molecular Backscatter - std')
    # (att_beta_mean + att_beta_std).plot(ax=ax, label='Attenuated  Backscatter + std')
    # ax.legend()
    # fig.savefig(f'residual_condition.png')
    # breakpoint()

    return mask


def get_mask_slope_attenuated_backscatter(
    slope_attenuated_backscatter: xr.DataArray,
    slope_attenuated_molecular_backscatter: xr.DataArray,
    relative_difference_thershold: float = 0.1,
) -> xr.DataArray:
    """Get mask for slope of attenuated backscatter within threshold limits.

    Args:
        slope_attenuated_backscatter (xr.DataArray): Slope of attenuated backscatter.
        slope_attenuated_molecular_backscatter (xr.DataArray): Slope of attenuated molecular backscatter.
        relative_difference_thershold (float, optional): Relative difference threshold. Defaults to 0.1.

    Returns:
        xr.DataArray: Mask for slope of attenuated backscatter.
    """
    # Get boolean variable where the slope of the attenuated backscatter is within the threshold limits
    mask = np.isclose(
        slope_attenuated_backscatter,
        slope_attenuated_molecular_backscatter,
        rtol=relative_difference_thershold,
    )

    mask = xr.DataArray(
        mask, dims=["range"], coords={"range": slope_attenuated_backscatter.range}
    )
    mask.attrs = {
        "long_name": "mask for slope of attenuated backscatter within threshold limits",
        "thresholds": f"diff < {relative_difference_thershold}",
        "units": "#",
    }
    return mask


def get_rayleigh_fit_analysis_data(
    channel: str,
    rf_dataset: xr.Dataset,
    initial_date: dt.datetime,
    final_date: dt.datetime,
    meteo_profiles: pd.DataFrame,
    reference_ranges: Tuple[float, float],
) -> xr.Dataset:
    """Rayleigh Fit for a given channel.

    Args:

        - channel (str): Channel to calculate the Rayleigh Fit.
        - rf_dataset (xr.Dataset): Dataset with the lidar measurements.
        - initial_date (dt.datetime): Initial date of the period to calculate the Rayleigh Fit.
        - final_date (dt.datetime): Final date of the period to calculate the Rayleigh Fit.
        - meteo_profiles (pd.DataFrame): Meteo profiles (e.g., from gfatpy.atmo.atmo.generate_meteo_profiles).
        - meteo_info (dict): Meteo information in a dictionary with keys: radiosonde_location, radiosonde_wmo_id, radiosonde_datetime.
        - reference_range (Tuple[float, float]): Reference range to normalize.
        - smooth_window (float): Range window to smooth the signal.

    Returns:

        - xr.Dataset: Rayleigh fit dataset.
    """

    wavelength, _, polarization, mode = channel2info(channel)
    ranges = rf_dataset.range.values
    # RCS
    signal = (
        rf_dataset[f"signal_{channel}"]
        .sel(time=slice(initial_date, final_date))
        .mean("time")
    )
    rcs = signal_to_rcs(signal, signal.range)

    snr = _snr(signal, reference_ranges)

    # Molecular properties from meteo profiles
    mol_properties = molecular_properties(
        wavelength, meteo_profiles["pressure"], meteo_profiles["temperature"], ranges
    )

    # Attenuated Molecular Backscatter
    attenuated_molecular_backscatter = mol_properties["atten_molecular_beta"]

    # Attenuated Backscatter
    attenuated_backscatter = _attenuated_backscatter(
        rcs, attenuated_molecular_backscatter, reference_ranges
    )

    # Attenuated Backscattering Ratio
    attenuated_backscattering_ratio = _attenuated_backscattering_ratio(
        attenuated_backscatter, attenuated_molecular_backscatter, reference_ranges
    )
    attenuated_backscattering_ratio_mask = get_mask_attenuated_backscattering_ratio(
        attenuated_backscattering_ratio
    )

    # Residual
    residual = _residual(attenuated_backscatter, attenuated_molecular_backscatter)
    residual_mask = get_mask_residual(
        attenuated_backscatter, attenuated_molecular_backscatter, reference_ranges
    )

    # Slope of Attenuated Backscatter
    slope_attenuated_backscatter = _slope_attenuated_backscatter(
        attenuated_backscatter, reference_ranges
    )

    # Slope of Attenuated Molecular Backscatter
    slope_attenuated_molecular_backscatter = _slope_attenuated_molecular_backscatter(
        attenuated_molecular_backscatter, reference_ranges
    )

    slope_mask = get_mask_slope_attenuated_backscatter(
        slope_attenuated_backscatter, slope_attenuated_molecular_backscatter
    )

    # Extinction
    extinction, std_extinction, durbin_watson_statistic = (
        _extinction_from_attenuated_backscatter(
            attenuated_backscatter, reference_ranges
        )
    )

    extinction_mask = get_mask_extinction(extinction, std_extinction)
    durbin_watson_mask = get_mask_durbin_watson(durbin_watson_statistic)

    # SNR
    snr = _snr(signal, reference_ranges)
    nsr_mask = get_mask_nsr(1 / snr)

    # Combine all masks
    all_mask = np.logical_and.reduce(
        [
            attenuated_backscattering_ratio_mask,
            residual_mask,
            slope_mask,
            extinction_mask,
            durbin_watson_mask,
            nsr_mask,
        ],
        axis=0,
    )

    all_mask = xr.DataArray(
        all_mask,
        dims=["range"],
        coords={"range": ranges},
        attrs={"long_name": "all masks applied", "units": "#"},
    )

    dataset = xr.Dataset(
        data_vars={
            "attenuated_backscatter": attenuated_backscatter,
            "attenuated_molecular_backscatter": attenuated_molecular_backscatter,
            "attenuated_backscattering_ratio": attenuated_backscattering_ratio,
            "residual": residual,
            "slope_attenuated_backscatter": slope_attenuated_backscatter,
            "slope_attenuated_molecular_backscatter": slope_attenuated_molecular_backscatter,
            "extinction": extinction,
            "std_extinction": std_extinction,
            "durbin_watson_statistic": durbin_watson_statistic,
            "snr": snr,
            "attenuated_backscattering_ratio_mask": attenuated_backscattering_ratio_mask,
            "residual_mask": residual_mask,
            "slope_mask": slope_mask,
            "extinction_mask": extinction_mask,
            "durbin_watson_mask": durbin_watson_mask,
            "nsr_mask": nsr_mask,
            "mask": all_mask,
            "minimum_range": reference_ranges[0],
            "maximum_range": reference_ranges[1],
        },
        coords={"range": ranges},
        attrs={
            "lidar_location": rf_dataset.attrs["location"],
            "lidar_name": rf_dataset.attrs["system"],
            "channel": channel,
            "datetime_ini": initial_date.strftime(
                "%Y-%m-%dT%H:%M:%S"
            ),  # FIXME: Dejar como estaba o pasar a ISO 8601?. Estaba en formato 20220808T12, ahora en 2022-08-08T12:00:00
            "datetime_end": final_date.strftime("%Y-%m-%dT%H:%M:%S"),
            "datetime_format": "%Y-%m-%dT%H:%M:%S",
            "duration": (final_date - initial_date).total_seconds(),
            "duration_units": "seconds",
        },
    )
    dataset["range"].attrs["units"] = "km"
    dataset["range"].attrs["long_name"] = "height"

    # Get optimum reference range
    # window_size_bin = _get_window_size(reference_ranges, dataset.range.values)
    # slope, _, _, msre, anderson_coefficient, _  = durbin_watson_test(
    # dataset.range.values,
    # dataset.attenuated_backscatter.values,
    # **{"length": window_size_bin, "degree": 1},
    # )

    # ydata = rolling_window_test(dataset.attenuated_backscatter.values, window_size_bin).T
    # abs_mean_attenuated_backscatter = np.abs(np.nanmean(ydata, axis=0))
    # std_attenuated_backscatter = np.nanstd(ydata, axis=0)
    # breakpoint()
    # weighting_function = np.abs(slope) * msre * anderson_coefficient.statistic
    # optimum_range = ranges[np.argmin(weighting_function)]
    # dataset['optimum_range'] = xr.DataArray(optimum_range, dims=[], attrs={"units": "km", "long_name": "optimum range"})

    #Save optimum range as the first range fulfilling the conditions in all_mask
    if all_mask.any():
        optimum_range = dataset.range.values[all_mask.values][0]
    else:
        optimum_range = -999.
        
    dataset["optimum_range"] = xr.DataArray(
        optimum_range,
        dims=[],
        attrs={"units": "m", "long_name": "optimum refenence range"},
    )
    return dataset


def plot_reference_analysis(dataset, **kwargs):
    fig, ax = plt.subplots(6, 1, figsize=kwargs.get("figsize", (15, 10)), sharex=True)

    dataset["range"] = dataset["range"] / 1e3

    # Plot attenuated total and molecular backscatter
    (1e6 * dataset["attenuated_molecular_backscatter"]).plot(
        x="range", ax=ax[0], label="$\\beta_{mol}^{att}$", color="blue"
    )
    (1e6 * dataset["attenuated_backscatter"]).plot(
        x="range", ax=ax[0], label="$\\beta^{att}$", color="red"
    )
    # Plot attenuated backscatter applying all_mask
    (1e6 * dataset["attenuated_backscatter"].where(dataset["mask"])).plot(
        x="range", ax=ax[0], color="green", label="$masked \\beta^{att}$"
    )
    ax[0].set_xlabel("Height (km)")
    ax[0].set_ylabel("$\\beta_{att}$\n$[Mm^{-1} sr^{-1}]$")
    ax[0].set_xticklabels([])

    # ax[1].set_yscale('log')  # Uncomment this line
    ax[0].set_ylim(0.1, 8)
    ax[0].set_yscale("log")
    ax[0].set_xlabel(None)  # Remove this line
    # ax[0].set_yscale("log")  # Uncomment this line

    # Plot attenuated backscattering ratio
    dataset["attenuated_backscattering_ratio"].plot(
        x="range", ax=ax[1], label="backscattering ratio", color="red"
    )
    dataset["attenuated_backscattering_ratio"].where(
        dataset["attenuated_backscattering_ratio_mask"]
    ).plot(x="range", ax=ax[1], color="green", label="$masked \\beta^{att}$")
    # Add horizontal line at 1
    ax[1].axhline(1, color="k", linestyle="--")
    ax[1].set_ylabel("$R_{att}$\n[#]")
    ax[1].set_ylim(0.8, 1.3)
    ax[1].set_xlabel(None)
    # ax[1].set_yscale('log')

    # Plot slopes
    (1e3 * dataset["slope_attenuated_molecular_backscatter"]).plot(
        x="range",
        ax=ax[2],
        color="k",
        label="slope of $\\beta_{mol}^{att}$",
    )
    (1e3 * dataset["slope_attenuated_backscatter"]).plot(
        x="range", ax=ax[2], color="red", label="slope of $\\beta^{att}$"
    )
    (1e3 * dataset["slope_attenuated_backscatter"]).where(dataset["slope_mask"]).plot(
        x="range", ax=ax[2], color="green", label="masked slope of $\\beta^{att}$"
    )
    ax[2].set_xlabel("")
    ax[2].set_ylabel("d$\\beta_{att}/dz$\n$[km^{-1} sr^{-1} m^{-1}]$")
    ax[2].set_ylim(-0.5, 0.5)

    # Plot extinctions
    (1e9 * dataset["extinction"]).plot(
        x="range", ax=ax[3], color="red", label="extinction"
    )
    (1e9 * dataset["std_extinction"]).plot(
        x="range", ax=ax[3], color="blue", label="std extinction"
    )
    (1e9 * dataset["extinction"]).where(dataset["extinction_mask"]).plot(
        x="range", ax=ax[3], color="green", label="masked extinction"
    )
    ax[3].axhline(0, color="k", linestyle="--")
    ax[3].set_xlabel("")
    ax[3].set_ylabel("$\\alpha$\n$[Gm^{-1}]$")
    ax[3].set_ylim(-0.1, 0.25)
    ax[3].set_xlim(*kwargs.get("range_limits", (0, 10)))
    # ax[3].set_xticklabels(np.arange(3, 11, 1))

    # Plot durbin-watson statistic
    dataset["durbin_watson_statistic"].plot(
        x="range", ax=ax[4], color="red", label="Durbin-Watson"
    )
    dataset["durbin_watson_statistic"].where(dataset["durbin_watson_mask"]).plot(
        x="range", ax=ax[4], color="green", label="masked Durbin-Watson"
    )
    ax[4].axhline(1, color="k", linestyle="--")
    ax[4].axhline(3, color="k", linestyle="--")
    ax[4].set_xlabel("")
    ax[4].set_ylabel("D-W stat.,\n\[#]")
    ax[4].set_ylim(0, 4)

    # Plot SNR
    (1 / dataset["snr"]).plot(x="range", ax=ax[5], color="red", label="NSR")
    (1 / dataset["snr"]).where(dataset["nsr_mask"]).plot(
        x="range", ax=ax[5], color="green", label="mask NSR"
    )

    ax[5].axhline(0.15, color="k", linestyle="--")
    ax[5].set_xlabel("Height (km)")
    ax[5].set_ylabel("NSR\n\[#]")
    ax[5].set_ylim(0, 0.25)
    range_limits = kwargs.get("range_limits", (0, 10000.0))
    ax[5].set_xlim(range_limits[0] / 1e3, range_limits[1] / 1e3)
    ax[5].set_xticklabels(ax[4].get_xticks())

    for ax_ in ax:
        ax_.legend(loc="upper left", fontsize=8)
        # Set fontsize of the labels to 8
        ax_.tick_params(axis="both", which="major", labelsize=8)
        ax_.minorticks_on()
        ax_.grid(which="both", linestyle="--", alpha=0.5)

    fig.savefig(
        kwargs.get("output_dir", Path.cwd())
        / f"attenuated_molecular_backscatter_vs_rcs2attbeta_{dataset.attrs['channel']}_{dataset['minimum_range'].item()}-{dataset['maximum_range'].item()}.png",
        dpi=kwargs.get("dpi", 300),
        bbox_inches="tight",
    )
    return fig
